'use strict';

var tslib = require('tslib');

/**
 * An error thrown if parsing failed. Custom checkers and transformers can throw this error to notify that the operation
 * has failed.
 */
var ValidationError = /** @class */ (function (_super) {
    tslib.__extends(ValidationError, _super);
    /**
     * Creates a new {@linkcode ValidationError} instance.
     *
     * @param issues The mutable list of partially defined issues that have caused an error.
     */
    function ValidationError(issues) {
        var _newTarget = this.constructor;
        var _this = this;
        var message = '';
        for (var _i = 0, issues_1 = issues; _i < issues_1.length; _i++) {
            var issue = issues_1[_i];
            inflateIssue(issue);
            message +=
                '\n' + issue.code + ' at /' + issue.path.join('/') + (issue.message != null ? ': ' + issue.message : '');
        }
        _this = _super.call(this, message) || this;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.name = 'ValidationError';
        _this.issues = issues;
        return _this;
    }
    return ValidationError;
}(Error));
function inflateIssue(issue) {
    if (issue.code == null) {
        issue.code = 'unknown';
    }
    if (issue.path == null) {
        issue.path = [];
    }
}

var objectTypes = ['object'];
var arrayTypes = ['array'];
var stringTypes = ['string'];
var numberTypes = ['number'];
var bigintTypes = ['bigint'];
var booleanTypes = ['boolean'];
var anyTypes = ['any'];
var neverTypes = ['never'];
function ok(value) {
    return { ok: true, value: value };
}
function unique(arr) {
    var uniqueArr = null;
    for (var i = 0; i < arr.length; ++i) {
        var value = arr[i];
        if (arr.includes(value, i + 1)) {
            if (uniqueArr === null) {
                uniqueArr = arr.slice(0, i);
            }
            continue;
        }
        if (uniqueArr !== null) {
            uniqueArr.push(value);
        }
    }
    return uniqueArr || arr;
}
var isArray = Array.isArray;
var getPrototypeOf = Object.getPrototypeOf;
function isEqual(a, b) {
    return a === b || (a !== a && b !== b);
}
function isObjectLike(value) {
    return value !== null && typeof value === 'object';
}
function isPlainObject(value) {
    var prototype;
    return isObjectLike(value) && ((prototype = getPrototypeOf(value)) === null || prototype.constructor === Object);
}
function isAsyncShapes(shapes) {
    var async = false;
    for (var i = 0; i < shapes.length && !async; ++i) {
        async = shapes[i].async;
    }
    return async;
}
/**
 * The convenient shortcut to add built-in checks to shapes.
 */
function appendCheck(shape, key, options, param, cb) {
    return shape.check(cb, {
        key: key,
        unsafe: options !== null && typeof options === 'object' && options.unsafe,
        param: param,
    });
}
function createIssueFactory(code, defaultMessage, options, param) {
    var paramKnown = arguments.length === 4;
    var meta;
    var message = defaultMessage;
    if (options !== null && typeof options === 'object') {
        if (options.message !== undefined) {
            message = options.message;
        }
        meta = options.meta;
    }
    else if (typeof options === 'function') {
        message = options;
    }
    else if (options != null) {
        message = options;
    }
    if (typeof message === 'function') {
        if (paramKnown) {
            return function (input, options) {
                return { code: code, path: [], input: input, message: message(param, code, input, meta, options), param: param, meta: meta };
            };
        }
        else {
            return function (input, options, param) {
                return { code: code, path: [], input: input, message: message(param, code, input, meta, options), param: param, meta: meta };
            };
        }
    }
    if (typeof message === 'string') {
        if (paramKnown) {
            message = message.replace('%s', param);
        }
        else if (message.indexOf('%s') !== -1) {
            return function (input, options, param) {
                return { code: code, path: [], input: input, message: message.replace('%s', param), param: param, meta: meta };
            };
        }
    }
    if (paramKnown) {
        return function (input) {
            return { code: code, path: [], input: input, message: message, param: param, meta: meta };
        };
    }
    else {
        return function (input, options, param) {
            return { code: code, path: [], input: input, message: message, param: param, meta: meta };
        };
    }
}
function unshiftPath(issues, key) {
    var issuesLength = issues.length;
    for (var i = 0; i < issuesLength; ++i) {
        issues[i].path.unshift(key);
    }
}
function concatIssues(issues, result) {
    if (issues !== null) {
        issues.push.apply(issues, result);
        return issues;
    }
    return result;
}
function pushIssue(issues, result) {
    if (issues !== null) {
        issues.push(result);
        return issues;
    }
    return [result];
}
function captureIssues(error) {
    if (error instanceof ValidationError) {
        return error.issues;
    }
    throw error;
}
function setFlag(flags, index) {
    if (typeof flags === 'number') {
        if (index < 32) {
            return flags | (1 << index);
        }
        flags = [flags, 0, 0];
    }
    flags[index >> 5] |= 1 << index % 32;
    return flags;
}
function isFlagSet(flag, index) {
    if (typeof flag === 'number') {
        return 0 !== flag >>> index;
    }
    else {
        return 0 !== flag[index >> 5] >>> index % 32;
    }
}
function setKeyValue(obj, key, value) {
    if (key === '__proto__') {
        Object.defineProperty(obj, key, { value: value, writable: true, enumerable: true, configurable: true });
    }
    else {
        obj[key] = value;
    }
}
function cloneEnumerableKeys(input, keyCount) {
    if (keyCount === void 0) { keyCount = -1; }
    var output = {};
    if (keyCount < 0) {
        for (var key in input) {
            setKeyValue(output, key, input[key]);
        }
    }
    if (keyCount > 0) {
        var index = 0;
        for (var key in input) {
            if (index === keyCount) {
                break;
            }
            setKeyValue(output, key, input[key]);
            ++index;
        }
    }
    return output;
}
function cloneKnownKeys(input, keys) {
    var output = {};
    var keysLength = keys.length;
    for (var i = 0; i < keysLength; ++i) {
        var key = keys[i];
        if (key in input) {
            setKeyValue(output, key, input[key]);
        }
    }
    return output;
}
function createApplyChecksCallback(checks) {
    var checksLength = checks.length;
    if (checksLength === 0) {
        return null;
    }
    if (checksLength === 1) {
        var _a = checks[0], unsafe0_1 = _a.unsafe, cb0_1 = _a.callback;
        return function (output, issues, options) {
            if (issues === null || unsafe0_1) {
                var result = void 0;
                try {
                    result = cb0_1(output, options);
                }
                catch (error) {
                    return concatIssues(issues, captureIssues(error));
                }
                if (result != null) {
                    return appendIssue(issues, result);
                }
            }
            return issues;
        };
    }
    if (checksLength === 2) {
        var _b = checks[0], unsafe0_2 = _b.unsafe, cb0_2 = _b.callback, _c = checks[1], unsafe1_1 = _c.unsafe, cb1_1 = _c.callback;
        return function (output, issues, options) {
            if (issues === null || unsafe0_2) {
                var result = void 0;
                try {
                    result = cb0_2(output, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                    if (!options.verbose) {
                        return issues;
                    }
                }
                if (result != null) {
                    issues = appendIssue(issues, result);
                    if (issues !== null && !options.verbose) {
                        return issues;
                    }
                }
            }
            if (issues === null || unsafe1_1) {
                var result = void 0;
                try {
                    result = cb1_1(output, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                }
                if (result != null) {
                    issues = appendIssue(issues, result);
                }
            }
            return issues;
        };
    }
    return function (output, issues, options) {
        for (var i = 0; i < checksLength; ++i) {
            var _a = checks[i], unsafe = _a.unsafe, callback = _a.callback;
            var result = void 0;
            if (issues !== null && !unsafe) {
                continue;
            }
            try {
                result = callback(output, options);
            }
            catch (error) {
                issues = concatIssues(issues, captureIssues(error));
                if (!options.verbose) {
                    return issues;
                }
            }
            if (result != null) {
                issues = appendIssue(issues, result);
                if (issues !== null && !options.verbose) {
                    return issues;
                }
            }
        }
        return issues;
    };
}
function appendIssue(issues, result) {
    if (isArray(result)) {
        var resultLength = result.length;
        if (resultLength === 0) {
            return issues;
        }
        for (var i = 0; i < resultLength; ++i) {
            inflateIssue(result[i]);
        }
        if (issues === null) {
            issues = result;
        }
        else {
            issues.push.apply(issues, result);
        }
    }
    else {
        inflateIssue(result);
        if (issues === null) {
            issues = [result];
        }
        else {
            issues.push(result);
        }
    }
    return issues;
}
function getInputTypes(shapes) {
    var types = [];
    for (var _i = 0, shapes_1 = shapes; _i < shapes_1.length; _i++) {
        var shape = shapes_1[_i];
        for (var _a = 0, _b = shape.inputTypes; _a < _b.length; _a++) {
            var type = _b[_a];
            types.push(type);
        }
    }
    return types;
}

var MESSAGE_ERROR_ASYNC = 'Shape is async, consider using tryAsync, parseAsync, or parseOrDefaultAsync';
var CODE_ARRAY_MAX = 'arrayMaxLength';
var CODE_ARRAY_MIN = 'arrayMinLength';
var CODE_ENUM = 'enum';
var CODE_EXCLUSION = 'exclusion';
var CODE_INSTANCE = 'instance';
var CODE_INTERSECTION = 'intersection';
var CODE_PREDICATE = 'predicate';
var CODE_NEVER = 'never';
var CODE_NUMBER_GT = 'numberGreaterThan';
var CODE_NUMBER_GTE = 'numberGreaterThanOrEqual';
var CODE_NUMBER_LT = 'numberLessThan';
var CODE_NUMBER_LTE = 'numberLessThanOrEqual';
var CODE_NUMBER_MULTIPLE_OF = 'numberMultipleOf';
var CODE_STRING_MAX = 'stringMaxLength';
var CODE_STRING_MIN = 'stringMinLength';
var CODE_STRING_REGEX = 'stringRegex';
var CODE_TYPE = 'type';
var CODE_TUPLE = 'tuple';
var CODE_UNION = 'union';
var CODE_UNKNOWN_KEYS = 'unknownKeys';
var TYPE_ARRAY = 'array';
var TYPE_BIGINT = 'bigint';
var TYPE_BOOLEAN = 'boolean';
var TYPE_INTEGER = 'integer';
var TYPE_NUMBER = 'number';
var TYPE_OBJECT = 'object';
var TYPE_PROMISE = 'promise';
var TYPE_STRING = 'string';
var MESSAGE_ARRAY_TYPE = 'Must be an array';
var MESSAGE_ARRAY_MAX = 'Must have the maximum length of %s';
var MESSAGE_ARRAY_MIN = 'Must have the minimum length of %s';
var MESSAGE_BIGINT_TYPE = 'Must be a bigint';
var MESSAGE_BOOLEAN_TYPE = 'Must be a boolean';
var MESSAGE_ENUM = 'Must be equal to one of: %s';
var MESSAGE_EXCLUSION = 'Must not be equal to %s';
var MESSAGE_INSTANCE = 'Must be a class instance';
var MESSAGE_INTEGER_TYPE = 'Must be an integer';
var MESSAGE_INTERSECTION = 'Must conform the intersection';
var MESSAGE_PREDICATE = 'Must conform the predicate';
var MESSAGE_NEVER = 'Must not be used';
var MESSAGE_NUMBER_TYPE = 'Must be a number';
var MESSAGE_NUMBER_GT = 'Must be greater than %s';
var MESSAGE_NUMBER_GTE = 'Must be greater than or equal to %s';
var MESSAGE_NUMBER_LT = 'Must be less than %s';
var MESSAGE_NUMBER_LTE = 'Must be less than or equal to %s';
var MESSAGE_NUMBER_MULTIPLE_OF = 'Must be a multiple of %s';
var MESSAGE_OBJECT_TYPE = 'Must be an object';
var MESSAGE_PROMISE_TYPE = 'Must be a promise';
var MESSAGE_UNKNOWN_KEYS = 'Must not have unknown keys: %s';
var MESSAGE_STRING_TYPE = 'Must be a string';
var MESSAGE_STRING_MAX = 'Must have the maximum length of %s';
var MESSAGE_STRING_MIN = 'Must have the minimum length of %s';
var MESSAGE_STRING_REGEX = 'Must match the pattern %s';
var MESSAGE_TUPLE = 'Must be a tuple of length %s';
var MESSAGE_UNION = 'Must conform a union';

var defaultParseOptions = Object.freeze({ verbose: false });
/**
 * The base shape.
 *
 * @template I The input value.
 * @template O The output value.
 */
var Shape = /** @class */ (function () {
    /**
     * Creates the new {@linkcode Shape} instance.
     *
     * @param inputTypes
     * @param async If `true` then the shape would allow only {@linkcode parseAsync} and throw an error if
     * {@linkcode parse} is called. Otherwise, shape can be used in both sync and async contexts.
     * @template I The input value.
     * @template O The output value.
     */
    function Shape(inputTypes, async) {
        if (async === void 0) { async = false; }
        /**
         * The list of checks applied to the shape output.
         */
        this.checks = [];
        /**
         * The human-readable shape description.
         */
        this.description = '';
        /**
         * Applies checks to the output.
         */
        this._applyChecks = null;
        /**
         * `true` if some checks from {@linkcode checks} were marked as unsafe, `false` otherwise.
         */
        this._unsafe = false;
        this.inputTypes =
            inputTypes === undefined || inputTypes.length === 0 || inputTypes.includes('any') ? anyTypes : unique(inputTypes);
        this.async = async;
        if (async) {
            this.apply = function () {
                throw new Error(MESSAGE_ERROR_ASYNC);
            };
        }
    }
    /**
     * Returns the type of the value.
     *
     * @param value The value to detect type of.
     */
    Shape.typeof = function (value) {
        var type = typeof value;
        if (type !== 'object') {
            return type;
        }
        if (value === null) {
            return 'null';
        }
        if (isArray(value)) {
            return 'array';
        }
        return type;
    };
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    Shape.prototype.at = function (key) {
        return null;
    };
    /**
     * Adds a human-readable description text to the shape.
     *
     * @param text The description text.
     * @returns The clone of this shape with the description added.
     */
    Shape.prototype.describe = function (text) {
        var shape = this._clone();
        shape.description = text;
        return shape;
    };
    /**
     * Appends the check that is applied to the shape output.
     *
     * If the {@linkcode CheckOptions.key} is defined and there's already a check with the same key then, it is removed
     * and a new check is appended to the list of shape checks. If the key is `undefined` then the `cb` identity is used
     * as a key.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The check options.
     * @returns The clone of this shape with the check added.
     */
    Shape.prototype.check = function (cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.key, key = _a === void 0 ? cb : _a, _b = options.unsafe, unsafe = _b === void 0 ? false : _b, param = options.param;
        var checks = this.checks.filter(function (check) { return check.key !== key; });
        checks.push({ key: key, callback: cb, unsafe: unsafe, param: param });
        var shape = this._clone();
        shape.checks = checks;
        shape._applyChecks = createApplyChecksCallback(checks);
        shape._unsafe || (shape._unsafe = unsafe);
        return shape;
    };
    /**
     * Redirects the shape output to another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @returns The {@linkcode RedirectShape} instance.
     * @template T The output value.
     */
    Shape.prototype.to = function (shape) {
        return new RedirectShape(this, shape);
    };
    /**
     * Synchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The output value.
     */
    Shape.prototype.transform = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, false, cb);
    };
    /**
     * Asynchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The transformed value.
     */
    Shape.prototype.transformAsync = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, true, cb);
    };
    Shape.prototype.refine = function (cb, options) {
        var issueFactory = createIssueFactory(CODE_PREDICATE, MESSAGE_PREDICATE, options, cb);
        return appendCheck(this, undefined, options, cb, function (input, options) {
            if (!cb(input)) {
                return issueFactory(input, options);
            }
        });
    };
    Shape.prototype.optional = function (defaultValue) {
        return new ReplaceShape(this, undefined, defaultValue);
    };
    Shape.prototype.nullable = function (defaultValue) {
        return new ReplaceShape(this, null, defaultValue);
    };
    Shape.prototype.nullish = function (defaultValue) {
        return this.nullable(defaultValue).optional(defaultValue);
    };
    /**
     * Prevents an input and output from being undefined.
     *
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     */
    Shape.prototype.nonOptional = function (options) {
        var shape = this;
        while (shape instanceof ReplaceShape && shape.checks.length === 0 && shape.searchedValue === undefined) {
            shape = shape.shape;
        }
        return new ExcludeShape(shape, undefined, options);
    };
    /**
     * Synchronously parses the input.
     *
     * Use {@linkcode parse} or {@linkcode try} instead of this method whenever possible. Override this method to
     * implement a custom shape.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    /**
     * Asynchronously parses the input.
     *
     * Use {@linkcode parseAsync} or {@linkcode tryAsync} instead of this method whenever possible. Override this method
     * to implement a custom shape that requires an async context.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype.applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) { return resolve(_this.apply(input, options)); });
    };
    /**
     * Returns the shallow clone of this shape.
     */
    Shape.prototype._clone = function () {
        return Object.assign(Object.create(getPrototypeOf(this)), this);
    };
    return Shape;
}());
Object.defineProperty(Shape.prototype, 'try', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return ok(input);
            }
            if (isArray(result)) {
                return { ok: false, issues: result };
            }
            return ok(result.value);
        };
        Object.defineProperty(this, 'try', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'tryAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return ok(input);
                }
                if (isArray(result)) {
                    return { ok: false, issues: result };
                }
                return result;
            });
        };
        Object.defineProperty(this, 'tryAsync', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parse', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return input;
            }
            if (isArray(result)) {
                throw new ValidationError(result);
            }
            return result.value;
        };
        Object.defineProperty(this, 'parse', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    throw new ValidationError(result);
                }
                return result.value;
            });
        };
        Object.defineProperty(this, 'parseAsync', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseOrDefault', {
    get: function () {
        var _this = this;
        var cb = function (input, defaultValue, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return input;
            }
            if (isArray(result)) {
                return defaultValue;
            }
            return result.value;
        };
        Object.defineProperty(this, 'parseOrDefault', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseOrDefaultAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, defaultValue, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    return defaultValue;
                }
                return result.value;
            });
        };
        Object.defineProperty(this, 'parseOrDefaultAsync', { value: cb });
        return cb;
    },
});
/**
 * The shape that applies a transformer callback to the base shape output.
 *
 * @template S The base shape.
 * @template O The transformed value.
 */
var TransformShape = /** @class */ (function (_super) {
    tslib.__extends(TransformShape, _super);
    /**
     * Creates the new {@linkcode TransformShape} instance.
     *
     * @param shape The base shape.
     * @param async If `true` then the transformed shape would await for the transformer to finish and use the resolved
     * value as an output. Otherwise, the value that is synchronously returned from the transformer is used as an output.
     * @param callback The callback that transforms the shape output value.
     * @template S The base shape.
     * @template O The transformed value.
     */
    function TransformShape(
    /**
     * The base shape which output value is transformed.
     */
    shape, async, 
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    callback) {
        var _this = _super.call(this, shape.inputTypes, shape.async || async) || this;
        _this.shape = shape;
        _this.callback = callback;
        return _this;
    }
    TransformShape.prototype.apply = function (input, options) {
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        var result = shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        try {
            output = callback(output, options);
        }
        catch (error) {
            return captureIssues(error);
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        if (isEqual(input, output)) {
            return null;
        }
        return ok(output);
    };
    TransformShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        return shape.applyAsync(input, options).then(function (result) {
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            return new Promise(function (resolve) { return resolve(callback(output, options)); }).then(function (output) {
                var issues;
                if (_applyChecks !== null) {
                    issues = _applyChecks(output, null, options);
                    if (issues !== null) {
                        return issues;
                    }
                }
                if (isEqual(input, output)) {
                    return null;
                }
                return ok(output);
            }, captureIssues);
        });
    };
    return TransformShape;
}(Shape));
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template I The input shape.
 * @template O The output shape.
 */
var RedirectShape = /** @class */ (function (_super) {
    tslib.__extends(RedirectShape, _super);
    /**
     * Creates the new {@linkcode RedirectShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template I The input shape.
     * @template O The output shape.
     */
    function RedirectShape(
    /**
     * The input shape.
     */
    inputShape, 
    /**
     * The output shape.
     */
    outputShape) {
        var _this = _super.call(this, inputShape.inputTypes, inputShape.async || outputShape.async) || this;
        _this.inputShape = inputShape;
        _this.outputShape = outputShape;
        return _this;
    }
    RedirectShape.prototype.apply = function (input, options) {
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        var result = inputShape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        var outputResult = outputShape.apply(output, options);
        if (outputResult !== null) {
            if (isArray(outputResult)) {
                return outputResult;
            }
            result = outputResult;
            output = outputResult.value;
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    RedirectShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var result = null;
        var output = input;
        return inputShape
            .applyAsync(input, options)
            .then(function (inputResult) {
            if (inputResult !== null) {
                if (isArray(inputResult)) {
                    return inputResult;
                }
                result = inputResult;
                output = inputResult.value;
            }
            return outputShape.applyAsync(output, options);
        })
            .then(function (outputResult) {
            var issues;
            if (outputResult !== null) {
                if (isArray(outputResult)) {
                    return outputResult;
                }
                result = outputResult;
                output = outputResult.value;
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return RedirectShape;
}(Shape));
/**
 * The shape that replaces an input value with another value.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template A The replaced value.
 * @template B The replacement value.
 */
var ReplaceShape = /** @class */ (function (_super) {
    tslib.__extends(ReplaceShape, _super);
    /**
     * Creates the new {@linkcode ReplaceShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param searchedValue The input that should be replaced.
     * @param value The replacement value. If `undefined` then the replaced value is returned as is without any parsing.
     * @template S The shape that parses the input without the replaced value.
     * @template A The searched value.
     * @template B The replacement value.
     */
    function ReplaceShape(
    /**
     * The shape that parses the input without the replaced value.
     */
    shape, 
    /**
     * The replaced value.
     */
    searchedValue, 
    /**
     * The replacement value.
     */
    value) {
        var _this = _super.call(this, shape.inputTypes.concat(Shape.typeof(searchedValue)), shape.async) || this;
        _this.shape = shape;
        _this.searchedValue = searchedValue;
        _this.value = value;
        _this._replacedResult = value === undefined || isEqual(value, searchedValue) ? null : ok(value);
        return _this;
    }
    ReplaceShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var issues;
        var output = input;
        var result = isEqual(input, this.searchedValue) ? this._replacedResult : this.shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    ReplaceShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _applyChecks = this._applyChecks;
        var issues;
        if (isEqual(input, this.searchedValue)) {
            if (_applyChecks !== null) {
                return new Promise(function (resolve) {
                    issues = _applyChecks(_this.value, null, options);
                    resolve(issues !== null ? issues : _this._replacedResult);
                });
            }
            return (this._replacedResultPromise || (this._replacedResultPromise = Promise.resolve(this._replacedResult)));
        }
        return this.shape.applyAsync(input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return ReplaceShape;
}(Shape));
/**
 * The shape that excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
var ExcludeShape = /** @class */ (function (_super) {
    tslib.__extends(ExcludeShape, _super);
    /**
     * Creates the new {@linkcode ExcludeShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param excludedValue The excluded value.
     * @param options The constraint options or an issue message.
     * @template S The shape that parses the input without the replaced value.
     * @template T The excluded value.
     */
    function ExcludeShape(
    /**
     * The base shape.
     */
    shape, 
    /**
     * The excluded value.
     */
    excludedValue, options) {
        var _this = _super.call(this, shape.inputTypes, shape.async) || this;
        _this.shape = shape;
        _this.excludedValue = excludedValue;
        _this._issueFactory = createIssueFactory(CODE_EXCLUSION, MESSAGE_EXCLUSION, options, excludedValue);
        return _this;
    }
    ExcludeShape.prototype.apply = function (input, options) {
        var _a = this, excludedValue = _a.excludedValue, _issueFactory = _a._issueFactory, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        if (isEqual(input, excludedValue)) {
            return [_issueFactory(input, options)];
        }
        var result = this.shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
            if (isEqual(output, excludedValue)) {
                return [_issueFactory(input, options)];
            }
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    ExcludeShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, excludedValue = _a.excludedValue, _issueFactory = _a._issueFactory, _applyChecks = _a._applyChecks;
        if (isEqual(input, excludedValue)) {
            return Promise.resolve([_issueFactory(input, options)]);
        }
        return this.shape.applyAsync(input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
                if (isEqual(output, excludedValue)) {
                    return [_issueFactory(input, options)];
                }
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return ExcludeShape;
}(Shape));

var integerRegex = /^(?:0|[1-9]\d*)$/;
/**
 * The shape of an array or a tuple.
 *
 * @template U The list of positioned element shapes or `null` if there are no positioned elements.
 * @template R The shape of rest elements or `null` if there are no rest elements.
 */
var ArrayShape = /** @class */ (function (_super) {
    tslib.__extends(ArrayShape, _super);
    /**
     * Creates a new {@linkcode ArrayShape} instance.
     *
     * @param shapes The list of positioned element shapes or `null` if there are no positioned elements.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The type constraint options or the type issue message.
     */
    function ArrayShape(
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    shapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape, options) {
        var _this = _super.call(this, arrayTypes, (shapes !== null && isAsyncShapes(shapes)) || (restShape !== null && restShape.async)) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._options = options;
        if (shapes !== null && restShape === null) {
            _this._issueFactory = createIssueFactory(CODE_TUPLE, MESSAGE_TUPLE, options, shapes.length);
        }
        else {
            _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_ARRAY_TYPE, options, TYPE_ARRAY);
        }
        return _this;
    }
    ArrayShape.prototype.at = function (key) {
        var shapes = this.shapes;
        var index = typeof key === 'number' ? key : typeof key !== 'string' ? -1 : integerRegex.test(key) ? parseInt(key, 10) : -1;
        if (index % 1 !== 0 || index < 0) {
            return null;
        }
        if (shapes !== null && index < shapes.length) {
            return shapes[index];
        }
        return this.restShape;
    };
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template T The shape of rest elements.
     */
    ArrayShape.prototype.rest = function (restShape) {
        return new ArrayShape(this.shapes, restShape, this._options);
    };
    /**
     * Constrains the array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MIN, MESSAGE_ARRAY_MIN, options, length);
        return appendCheck(this, CODE_ARRAY_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MAX, MESSAGE_ARRAY_MAX, options, length);
        return appendCheck(this, CODE_ARRAY_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    ArrayShape.prototype.apply = function (input, options) {
        var _a = this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var inputLength;
        var shapesLength = 0;
        // noinspection CommaExpressionJS
        if (!isArray(input) ||
            ((inputLength = input.length),
                shapes !== null && inputLength !== (shapesLength = shapes.length) && restShape === null)) {
            return [this._issueFactory(input, options)];
        }
        var issues = null;
        var output = input;
        if (shapes !== null || restShape !== null) {
            for (var i = 0; i < inputLength; ++i) {
                var value = input[i];
                var valueShape = i < shapesLength ? shapes[i] : restShape;
                var result = valueShape.apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, i);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = input.slice(0);
                    }
                    output[i] = result.value;
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    ArrayShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            var _a = _this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var inputLength;
            var shapesLength = 0;
            // noinspection CommaExpressionJS
            if (!isArray(input) ||
                ((inputLength = input.length),
                    shapes !== null && inputLength !== (shapesLength = shapes.length) && restShape === null)) {
                resolve([_this._issueFactory(input, options)]);
                return;
            }
            var promises = [];
            if (shapes !== null || restShape !== null) {
                for (var i = 0; i < inputLength; ++i) {
                    var value = input[i];
                    var valueShape = i < shapesLength ? shapes[i] : restShape;
                    promises.push(valueShape.applyAsync(value, options));
                }
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var issues = null;
                var output = input;
                for (var i = 0; i < resultsLength; ++i) {
                    var result = results[i];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, i);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[i], result.value)) {
                        if (input === output) {
                            output = input.slice(0);
                        }
                        output[i] = result.value;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    return ArrayShape;
}(Shape));

/**
 * The shape of the bigint value.
 */
var BigIntShape = /** @class */ (function (_super) {
    tslib.__extends(BigIntShape, _super);
    /**
     * Creates a new {@linkcode BigIntShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function BigIntShape(options) {
        var _this = _super.call(this, bigintTypes) || this;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_BIGINT_TYPE, options, TYPE_BIGINT);
        return _this;
    }
    BigIntShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (typeof input !== 'bigint') {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return BigIntShape;
}(Shape));

/**
 * The shape of the bigint value.
 */
var BooleanShape = /** @class */ (function (_super) {
    tslib.__extends(BooleanShape, _super);
    /**
     * Creates a new {@linkcode BooleanShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function BooleanShape(options) {
        var _this = _super.call(this, booleanTypes) || this;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_BOOLEAN_TYPE, options, TYPE_BOOLEAN);
        return _this;
    }
    BooleanShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (typeof input !== 'boolean') {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return BooleanShape;
}(Shape));

/**
 * The shape that constrains an input to one of values.
 *
 * @template T Allowed values.
 */
var EnumShape = /** @class */ (function (_super) {
    tslib.__extends(EnumShape, _super);
    /**
     * Creates a new {@linkcode EnumShape} instance.
     *
     * @param values The list of values allowed for the input.
     * @param options The type constraint options or an issue message.
     */
    function EnumShape(values, options) {
        var _this = _super.call(this, values.map(Shape.typeof)) || this;
        _this.values = values;
        _this._issueFactory = createIssueFactory(CODE_ENUM, MESSAGE_ENUM, options, values);
        return _this;
    }
    EnumShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (!this.values.includes(input)) {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return EnumShape;
}(Shape));

/**
 * The shape of the class instance.
 *
 * @template C The class constructor.
 */
var InstanceShape = /** @class */ (function (_super) {
    tslib.__extends(InstanceShape, _super);
    /**
     * Creates a new {@linkcode InstanceShape} instance.
     *
     * @param ctor The class constructor.
     * @param options The type constraint options or the type issue message.
     * @template C The class constructor.
     */
    function InstanceShape(ctor, options) {
        var _this = _super.call(this, ctor === Array || Array.prototype.isPrototypeOf(ctor.prototype) ? arrayTypes : objectTypes) || this;
        _this.ctor = ctor;
        _this._issueFactory = createIssueFactory(CODE_INSTANCE, MESSAGE_INSTANCE, options, ctor);
        return _this;
    }
    InstanceShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (!(input instanceof this.ctor)) {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return InstanceShape;
}(Shape));

var IntersectionShape = /** @class */ (function (_super) {
    tslib.__extends(IntersectionShape, _super);
    function IntersectionShape(shapes, options) {
        var _this = _super.call(this, [], isAsyncShapes(shapes)) || this;
        _this.shapes = shapes;
        _this._issueFactory = createIssueFactory(CODE_INTERSECTION, MESSAGE_INTERSECTION, options, undefined);
        return _this;
    }
    IntersectionShape.prototype.apply = function (input, options) {
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var outputs = null;
        for (var i = 0; i < shapesLength; ++i) {
            var result = shapes[i].apply(input, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                return result;
            }
            var output = result.value;
            if (outputs === null) {
                outputs = [output];
                continue;
            }
            if (outputs.includes(output)) {
                continue;
            }
            outputs.push(output);
        }
        if (outputs === null) {
            return null;
        }
        if (outputs.length !== shapesLength) {
            outputs.push(input);
        }
        return intersectOutputs(input, outputs, this._issueFactory, options);
    };
    IntersectionShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var promises = [];
        for (var i = 0; i < shapesLength; ++i) {
            promises.push(shapes[i].applyAsync(input, options));
        }
        return Promise.all(promises).then(function (results) {
            var outputs = null;
            for (var i = 0; i < shapesLength; ++i) {
                var result = results[i];
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    return result;
                }
                var output = result.value;
                if (outputs === null) {
                    outputs = [output];
                    continue;
                }
                if (outputs.includes(output)) {
                    continue;
                }
                outputs.push(output);
            }
            if (outputs === null) {
                return null;
            }
            if (outputs.length !== shapesLength) {
                outputs.push(input);
            }
            return intersectOutputs(input, outputs, _this._issueFactory, options);
        });
    };
    return IntersectionShape;
}(Shape));
var NEVER = Symbol();
function intersectOutputs(input, outputs, issueFactory, options) {
    var outputsLength = outputs.length;
    var value = outputs[0];
    for (var i = 1; i < outputsLength; ++i) {
        value = intersectPair(value, outputs[i]);
        if (value === NEVER) {
            return [issueFactory(input, options)];
        }
    }
    return ok(value);
}
function intersectPair(a, b) {
    if (isEqual(a, b)) {
        return a;
    }
    var aType = Shape.typeof(a);
    var bType = Shape.typeof(b);
    if (aType !== bType) {
        return NEVER;
    }
    if (aType === 'object') {
        var output = Object.assign({}, a, b);
        for (var key in a) {
            if (key in b) {
                var outputValue = intersectPair(a[key], b[key]);
                if (outputValue === NEVER) {
                    return NEVER;
                }
                output[key] = outputValue;
            }
        }
        return output;
    }
    if (aType === 'array') {
        var aLength = a.length;
        if (aLength === b.length) {
            var output = a;
            for (var i = 0; i < aLength; ++i) {
                var aValue = a[i];
                var bValue = b[i];
                if (isEqual(aValue, bValue)) {
                    continue;
                }
                if (output === a) {
                    output = a.slice(0);
                }
                var outputValue = intersectPair(aValue, bValue);
                if (outputValue === NEVER) {
                    return NEVER;
                }
                output[i] = outputValue;
            }
            return output;
        }
    }
    return NEVER;
}

/**
 * Lazily resolves a shape using the provider.
 *
 * @template S The base shape.
 */
var LazyShape = /** @class */ (function (_super) {
    tslib.__extends(LazyShape, _super);
    /**
     * Creates a new {@linkcode LazyShape} instance.
     *
     * @param provider The provider that returns the base shape.
     * @param async If `true` the shape would support both sync and async base shapes, otherwise only sync shapes are
     * allowed.
     * @template S The base shape.
     */
    function LazyShape(provider, async) {
        var _this = _super.call(this, anyTypes, async) || this;
        _this._provider = provider;
        return _this;
    }
    LazyShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues;
        var result = this.shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    LazyShape.prototype.applyAsync = function (input, options) {
        var _applyChecks = this._applyChecks;
        return this.shape.applyAsync(input, options).then(function (result) {
            var output = input;
            var issues;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return LazyShape;
}(Shape));
Object.defineProperty(LazyShape.prototype, 'shape', {
    get: function () {
        var shape = this._provider();
        Object.defineProperty(this, 'shape', { value: shape, enumerable: true });
        return shape;
    },
});

/**
 * The shape that doesn't match any input.
 */
var NeverShape = /** @class */ (function (_super) {
    tslib.__extends(NeverShape, _super);
    /**
     * Creates a new {@linkcode NeverShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function NeverShape(options) {
        var _this = _super.call(this, neverTypes) || this;
        _this._issueFactory = createIssueFactory(CODE_NEVER, MESSAGE_NEVER, options, undefined);
        return _this;
    }
    NeverShape.prototype.apply = function (input, options) {
        return [this._issueFactory(input, options)];
    };
    return NeverShape;
}(Shape));

/**
 * The shape of the finite number.
 */
var NumberShape = /** @class */ (function (_super) {
    tslib.__extends(NumberShape, _super);
    /**
     * Creates a new {@linkcode NumberShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function NumberShape(options) {
        var _this = _super.call(this, numberTypes) || this;
        _this._typePredicate = Number.isFinite;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_NUMBER_TYPE, options, TYPE_NUMBER);
        return _this;
    }
    /**
     * Constrains the number to be greater than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.positive = function (options) {
        return this.gt(0, options);
    };
    /**
     * Constrains the number to be less than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.negative = function (options) {
        return this.lt(0, options);
    };
    /**
     * Constrains the number to be greater than the value.
     *
     * @param value The exclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GT, MESSAGE_NUMBER_GT, options, value);
        return appendCheck(this, CODE_NUMBER_GT, options, value, function (input, options) {
            if (input <= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than the value.
     *
     * @param value The exclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LT, MESSAGE_NUMBER_LT, options, value);
        return appendCheck(this, CODE_NUMBER_LT, options, value, function (input, options) {
            if (input >= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GTE, MESSAGE_NUMBER_GTE, options, value);
        return appendCheck(this, CODE_NUMBER_GTE, options, value, function (input, options) {
            if (input < value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LTE, MESSAGE_NUMBER_LTE, options, value);
        return appendCheck(this, CODE_NUMBER_LTE, options, value, function (input, options) {
            if (input > value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param value The number by which the input should be divisible without a remainder.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.multipleOf = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_MULTIPLE_OF, MESSAGE_NUMBER_MULTIPLE_OF, options, value);
        return appendCheck(this, CODE_NUMBER_MULTIPLE_OF, options, value, function (input, options) {
            if (input % value !== 0) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be an integer and rejects `Infinity` and `NaN` values.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.integer = function (options) {
        var shape = this._clone();
        shape._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_INTEGER_TYPE, options, TYPE_INTEGER);
        shape._typePredicate = Number.isInteger;
        return shape;
    };
    NumberShape.prototype.apply = function (input, options) {
        var _a = this, _typePredicate = _a._typePredicate, _applyChecks = _a._applyChecks;
        if (!_typePredicate(input)) {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return NumberShape;
}(Shape));

/**
 * The shape of an object.
 *
 * @template P The mapping from an object key to a corresponding value shape.
 * @template R The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
 */
var ObjectShape = /** @class */ (function (_super) {
    tslib.__extends(ObjectShape, _super);
    /**
     * Creates a new {@linkcode ObjectShape} instance.
     *
     * @param shapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The type constraint options or an issue message.
     * @param keysMode The mode of unknown keys handling.
     * @template P The mapping from an object key to a corresponding value shape.
     * @template R The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    function ObjectShape(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    shapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape, options, keysMode) {
        if (keysMode === void 0) { keysMode = 'preserved'; }
        var _this = this;
        var valueShapes = Object.values(shapes);
        _this = _super.call(this, objectTypes, isAsyncShapes(valueShapes) || (restShape !== null && restShape.async)) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._typePredicate = isObjectLike;
        _this.keys = Object.keys(shapes);
        _this.keysMode = keysMode;
        _this._options = options;
        _this._valueShapes = valueShapes;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    ObjectShape.prototype.at = function (key) {
        return this.shapes.hasOwnProperty(key) ? this.shapes[key] : this.restShape;
    };
    ObjectShape.prototype.extend = function (shape) {
        var shapes = Object.assign({}, this.shapes, shape instanceof ObjectShape ? shape.shapes : shape);
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    ObjectShape.prototype.pick = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    ObjectShape.prototype.omit = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (!keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.partial = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.optional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.required = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.nonOptional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * The returned object shape would have no checks.
     *
     * @param options The constraint options or an issue message.
     * @returns The new object shape.
     */
    ObjectShape.prototype.exact = function (options) {
        var shape = new ObjectShape(this.shapes, null, this._options, 'exact');
        shape._exactIssueFactory = createIssueFactory(CODE_UNKNOWN_KEYS, MESSAGE_UNKNOWN_KEYS, options);
        return shape;
    };
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.strip = function () {
        return new ObjectShape(this.shapes, null, this._options, 'stripped');
    };
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.preserve = function () {
        return new ObjectShape(this.shapes, null, this._options);
    };
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape that validates values at unknown keys.
     * @returns The new object shape.
     * @template T The index signature shape.
     */
    ObjectShape.prototype.rest = function (restShape) {
        return new ObjectShape(this.shapes, restShape, this._options);
    };
    /**
     * Returns the enum shape of keys of this object.
     */
    ObjectShape.prototype.keyof = function () {
        return new EnumShape(this.keys);
    };
    /**
     * Constrains an object to be an `Object` instance or to have a `null` prototype.
     */
    ObjectShape.prototype.plain = function () {
        var shape = this._clone();
        shape._typePredicate = isPlainObject;
        return shape;
    };
    ObjectShape.prototype.apply = function (input, options) {
        var _typePredicate = this._typePredicate;
        if (!_typePredicate(input)) {
            return [this._typeIssueFactory(input, options)];
        }
        if (this.keysMode === 'preserved' && this.restShape === null) {
            return this._applyLax(input, options);
        }
        else {
            return this._applyStrict(input, options);
        }
    };
    ObjectShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            var _typePredicate = _this._typePredicate;
            if (!_typePredicate(input)) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var _a = _this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var keysLength = keys.length;
            var promises = [];
            var issues = null;
            var output = input;
            var seenCount = 0;
            var seenFlags = 0;
            var unknownKeys = null;
            for (var key in input) {
                var value = input[key];
                var index = keys.indexOf(key);
                var valueShape = restShape;
                if (index !== -1) {
                    seenCount++;
                    seenFlags = setFlag(seenFlags, index);
                    valueShape = _valueShapes[index];
                }
                if (valueShape !== null) {
                    promises.push(key, valueShape.applyAsync(value, options));
                    continue;
                }
                if (keysMode === 'exact') {
                    if (unknownKeys !== null) {
                        unknownKeys.push(key);
                        continue;
                    }
                    unknownKeys = [key];
                    if (!options.verbose) {
                        break;
                    }
                    continue;
                }
                if (input === output && keysMode === 'stripped') {
                    output = cloneKnownKeys(input, keys);
                }
            }
            if (unknownKeys !== null) {
                var issue = _this._exactIssueFactory(input, options, unknownKeys);
                if (!options.verbose) {
                    resolve([issue]);
                    return;
                }
                issues = pushIssue(issues, issue);
            }
            if (seenCount !== keysLength) {
                for (var i = 0; i < keysLength; ++i) {
                    if (isFlagSet(seenFlags, i)) {
                        continue;
                    }
                    var key = keys[i];
                    var value = input[key];
                    promises.push(key, _valueShapes[i].applyAsync(value, options));
                }
            }
            resolve(Promise.all(promises).then(function (entries) {
                var entriesLength = entries.length;
                for (var i = 0; i < entriesLength; i += 2) {
                    var key = entries[i];
                    var result = entries[i + 1];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, key);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[key], result.value)) {
                        if (input === output) {
                            output = cloneEnumerableKeys(input);
                        }
                        setKeyValue(output, key, result.value);
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    ObjectShape.prototype._applyLax = function (input, options) {
        var _a = this, keys = _a.keys, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        for (var i = 0; i < keysLength; ++i) {
            var key = keys[i];
            var value = input[key];
            var result = _valueShapes[i].apply(value, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftPath(result, key);
                if (!options.verbose) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                if (input === output) {
                    output = cloneEnumerableKeys(input);
                }
                setKeyValue(output, key, result.value);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Unknown keys are either parsed with a {@linkcode restShape}, stripped, or cause an issue.
     */
    ObjectShape.prototype._applyStrict = function (input, options) {
        var _a = this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        var seenCount = 0;
        var seenFlags = 0;
        var unknownKeys = null;
        for (var key in input) {
            var value = input[key];
            var index = keys.indexOf(key);
            var valueShape = restShape;
            // The key is known
            if (index !== -1) {
                seenCount++;
                seenFlags = setFlag(seenFlags, index);
                valueShape = _valueShapes[index];
            }
            // The key is known or indexed
            if (valueShape !== null) {
                var result = valueShape.apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = restShape === null ? cloneKnownKeys(input, keys) : cloneEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
                continue;
            }
            // Unknown keys raise an issue
            if (keysMode === 'exact') {
                if (unknownKeys !== null) {
                    unknownKeys.push(key);
                    continue;
                }
                unknownKeys = [key];
                if (!options.verbose) {
                    break;
                }
                continue;
            }
            // Unknown keys are stripped
            if (input === output && (_unsafe || issues === null)) {
                output = cloneKnownKeys(input, keys);
            }
        }
        // Raise unknown keys issue
        if (unknownKeys !== null) {
            var issue = this._exactIssueFactory(input, options, unknownKeys);
            if (!options.verbose) {
                return [issue];
            }
            issues = pushIssue(issues, issue);
        }
        // Parse absent known keys
        if (seenCount !== keysLength) {
            for (var i = 0; i < keysLength; ++i) {
                if (isFlagSet(seenFlags, i)) {
                    continue;
                }
                var key = keys[i];
                var value = input[key];
                var result = _valueShapes[i].apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = cloneEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    return ObjectShape;
}(Shape));

/**
 * The shape of a value wrapped in a `Promise` instance.
 *
 * @template S The shape of the resolved value.
 */
var PromiseShape = /** @class */ (function (_super) {
    tslib.__extends(PromiseShape, _super);
    /**
     * Creates a new {@linkcode PromiseShape} instance.
     *
     * @param shape The shape of the resolved value.
     * @param options The type constraint options or the type issue message.
     * @template S The shape of the resolved value.
     */
    function PromiseShape(shape, options) {
        var _this = _super.call(this, objectTypes, true) || this;
        _this.shape = shape;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_PROMISE_TYPE, options, TYPE_PROMISE);
        return _this;
    }
    PromiseShape.prototype.apply = function (input, options) {
        throw new Error(MESSAGE_ERROR_ASYNC);
    };
    PromiseShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!(input instanceof Promise)) {
            return Promise.resolve([this._issueFactory(input, options)]);
        }
        var _applyChecks = this._applyChecks;
        var inputValue;
        var outputValue;
        return input
            .then(function (value) {
            inputValue = outputValue = value;
            return _this.shape.applyAsync(value, options);
        })
            .then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                outputValue = result.value;
                if (!isEqual(inputValue, outputValue)) {
                    output = Promise.resolve(outputValue);
                }
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return ok(output);
        });
    };
    return PromiseShape;
}(Shape));

/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
var RecordShape = /** @class */ (function (_super) {
    tslib.__extends(RecordShape, _super);
    /**
     * Creates a new {@linkcode RecordShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape or `null` if keys should be preserved intact.
     * @template V The value shape.
     */
    function RecordShape(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this, objectTypes, (keyShape != null && keyShape.async) || valueShape.async) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    RecordShape.prototype.at = function (key) {
        return typeof key === 'string' || typeof key === 'number' ? this.valueShape : null;
    };
    RecordShape.prototype.apply = function (input, options) {
        if (!isObjectLike(input)) {
            return [this._issueFactory(input, options)];
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keyCount = 0;
        var issues = null;
        var output = input;
        for (var key in input) {
            var value = input[key];
            var outputKey = key;
            var outputValue = value;
            if (keyShape !== null) {
                var keyResult = keyShape.apply(key, options);
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftPath(keyResult, key);
                        if (!options.verbose) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                    }
                    else {
                        outputKey = keyResult.value;
                    }
                }
            }
            var valueResult = valueShape.apply(value, options);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftPath(valueResult, key);
                    if (!options.verbose) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                }
                else {
                    outputValue = valueResult.value;
                }
            }
            if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                if (input === output) {
                    output = cloneEnumerableKeys(input, keyCount);
                }
                setKeyValue(output, outputKey, outputValue);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    RecordShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            if (!isObjectLike(input)) {
                resolve([_this._issueFactory(input, options)]);
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var promises = [];
            for (var key in input) {
                var value = input[key];
                promises.push(key, keyShape !== null ? keyShape.applyAsync(key, options) : null, valueShape.applyAsync(value, options));
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var keyCount = 0;
                var issues = null;
                var output = input;
                for (var i = 0; i < resultsLength; i += 3) {
                    var key = results[i];
                    var value = input[key];
                    var outputKey = key;
                    var outputValue = value;
                    var keyResult = results[i + 1];
                    var valueResult = results[i + 2];
                    if (keyResult !== null) {
                        if (isArray(keyResult)) {
                            unshiftPath(keyResult, key);
                            if (!options.verbose) {
                                return keyResult;
                            }
                            issues = concatIssues(issues, keyResult);
                        }
                        else {
                            outputKey = keyResult.value;
                        }
                    }
                    if (valueResult !== null) {
                        if (isArray(valueResult)) {
                            unshiftPath(valueResult, key);
                            if (!options.verbose) {
                                return valueResult;
                            }
                            issues = concatIssues(issues, valueResult);
                        }
                        else {
                            outputValue = valueResult.value;
                        }
                    }
                    if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                        if (input === output) {
                            output = cloneEnumerableKeys(input, keyCount);
                        }
                        output[outputKey] = outputValue;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    return RecordShape;
}(Shape));

/**
 * The shape that constrains the input as a string.
 */
var StringShape = /** @class */ (function (_super) {
    tslib.__extends(StringShape, _super);
    /**
     * Creates a new {@linkcode StringShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function StringShape(options) {
        var _this = _super.call(this, stringTypes) || this;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_STRING_TYPE, options, TYPE_STRING);
        return _this;
    }
    /**
     * The shortcut to apply both {@linkcode min} and {@linkcode max} constraints.
     *
     * @param length The exact length a string must have.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the string length to be greater than or equal to the length.
     *
     * @param length The minimum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MIN, MESSAGE_STRING_MIN, options, length);
        return appendCheck(this, CODE_STRING_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string length to be less than or equal to the length.
     *
     * @param length The maximum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MAX, MESSAGE_STRING_MAX, options, length);
        return appendCheck(this, CODE_STRING_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string to match a regexp.
     *
     * @param re The regular expression that the sting must conform.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.regex = function (re, options) {
        var issueFactory = createIssueFactory(CODE_STRING_REGEX, MESSAGE_STRING_REGEX, options, re);
        return appendCheck(this, CODE_STRING_REGEX, options, re, function (input, options) {
            re.lastIndex = 0;
            if (!re.test(input)) {
                return issueFactory(input, options);
            }
        });
    };
    StringShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (typeof input !== 'string') {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return StringShape;
}(Shape));

/**
 * The shape that requires an input to conform at least one of the united shapes.
 *
 * @template U The list of united shapes.
 */
var UnionShape = /** @class */ (function (_super) {
    tslib.__extends(UnionShape, _super);
    /**
     * Creates a new {@linkcode UnionShape} instance.
     *
     * @param shapes The list of united shapes.
     * @param options The union constraint options or an issue message.
     * @template U The list of united shapes.
     */
    function UnionShape(
    /**
     * The list of united shapes.
     */
    shapes, options) {
        var _this = _super.call(this, getInputTypes(shapes), isAsyncShapes(shapes)) || this;
        _this.shapes = shapes;
        var _a = createUnionBuckets(shapes), buckets = _a.buckets, anyBucket = _a.anyBucket;
        _this._buckets = buckets;
        _this._anyBucket = anyBucket;
        _this._issueFactory = createIssueFactory(CODE_UNION, MESSAGE_UNION, options);
        return _this;
    }
    UnionShape.prototype.at = function (key) {
        var valueShapes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var valueShape = shape.at(key);
            if (valueShape !== null) {
                valueShapes.push(valueShape);
            }
        }
        if (valueShapes.length === 0) {
            return null;
        }
        if (valueShapes.length === 1) {
            return valueShapes[0];
        }
        return new UnionShape(valueShapes);
    };
    UnionShape.prototype.apply = function (input, options) {
        var _a = this, _buckets = _a._buckets, _anyBucket = _a._anyBucket, _applyChecks = _a._applyChecks;
        var bucket = _buckets !== null ? _buckets[Shape.typeof(input)] || _anyBucket : _anyBucket;
        var issues = null;
        var result = null;
        var output = input;
        var bucketLength = 0;
        var index = 0;
        if (bucket !== null) {
            for (bucketLength = bucket.length; index < bucketLength; ++index) {
                result = bucket[index].apply(input, options);
                if (result === null) {
                    break;
                }
                if (isArray(result)) {
                    issues = concatIssues(issues, result);
                    continue;
                }
                output = result.value;
                break;
            }
        }
        if (index === bucketLength) {
            return [this._issueFactory(input, options, issues)];
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    UnionShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, _buckets = _a._buckets, _anyBucket = _a._anyBucket, _applyChecks = _a._applyChecks;
        var bucket = _buckets !== null ? _buckets[Shape.typeof(input)] || _anyBucket : _anyBucket;
        if (bucket === null) {
            return Promise.resolve([this._issueFactory(input, options, [])]);
        }
        var bucketLength = bucket.length;
        var issues = null;
        var index = 0;
        var nextShape = function () {
            return bucket[index].applyAsync(input, options).then(function (result) {
                ++index;
                var output = input;
                if (result !== null) {
                    if (isArray(result)) {
                        issues = concatIssues(issues, result);
                        if (index === bucketLength) {
                            return [_this._issueFactory(input, options, issues)];
                        }
                        return nextShape();
                    }
                    output = result.value;
                }
                if (_applyChecks !== null) {
                    issues = _applyChecks(output, null, options);
                    if (issues !== null) {
                        return issues;
                    }
                }
                return result;
            });
        };
        return nextShape();
    };
    return UnionShape;
}(Shape));
/**
 * Creates a mapping from the value type to an array of shapes that are applicable.
 *
 * @param shapes The list of united shapes.
 */
function createUnionBuckets(shapes) {
    var buckets = {};
    var anyBucket = null;
    var bucketTypes = [];
    for (var _i = 0, _a = unwrapUnionShapes(shapes); _i < _a.length; _i++) {
        var shape = _a[_i];
        // Collect shapes that can parse any input
        if (shape.inputTypes.includes('any')) {
            anyBucket || (anyBucket = []);
            if (!anyBucket.includes(shape)) {
                anyBucket.push(shape);
            }
            continue;
        }
        // Populate buckets that require specific input types
        for (var _b = 0, _c = shape.inputTypes; _b < _c.length; _b++) {
            var type = _c[_b];
            var bucket = buckets[type];
            if (!bucket) {
                bucketTypes.push(type);
                buckets[type] = [shape];
                continue;
            }
            if (!bucket.includes(shape)) {
                bucket.push(shape);
            }
        }
    }
    if (bucketTypes.length === 0) {
        buckets = null;
    }
    else if (anyBucket !== null && buckets !== null) {
        for (var _d = 0, bucketTypes_1 = bucketTypes; _d < bucketTypes_1.length; _d++) {
            var type = bucketTypes_1[_d];
            buckets[type] = unique(buckets[type].concat(anyBucket));
        }
    }
    return { buckets: buckets, anyBucket: anyBucket };
}
/**
 * Unwraps nested union shapes that don't have any checks.
 */
function unwrapUnionShapes(opaqueShapes) {
    var shapes = [];
    for (var _i = 0, opaqueShapes_1 = opaqueShapes; _i < opaqueShapes_1.length; _i++) {
        var shape = opaqueShapes_1[_i];
        if (shape instanceof UnionShape && shape.checks.length === 0) {
            shapes.push.apply(shapes, unwrapUnionShapes(shape.shapes));
        }
        else {
            shapes.push(shape);
        }
    }
    return shapes;
}

function any(cb, options) {
    var shape = new Shape();
    return cb != null ? shape.refine(cb, options) : shape;
}

function array(shape, options) {
    if (shape == null || shape instanceof Shape) {
        return new ArrayShape(null, shape || null, options);
    }
    else {
        return new ArrayShape(null, null, shape);
    }
}

/**
 * Creates the bigint shape.
 *
 * @param options The constraint options or an issue message.
 */
function bigint(options) {
    return new BigIntShape(options);
}

/**
 * Creates the boolean shape.
 *
 * @param options The constraint options or an issue message.
 */
function boolean(options) {
    return new BooleanShape(options);
}

function enum_(values, options) {
    return new EnumShape(unique(isArray(values) ? values : Object.values(values).filter(function (value) { return typeof values[value] !== 'number'; })), options);
}

/**
 * Creates the class instance shape.
 *
 * @param ctor The instance constructor.
 * @param options The constraint options or an issue message.
 * @template C The instance constructor.
 */
function instanceOf(ctor, options) {
    return new InstanceShape(ctor, options);
}

/**
 * Creates an intersection shape that tries to parse the input with all provided shapes.
 *
 * @param shapes The list of shapes.
 * @param options The constraint options or an issue message.
 * @template U The tuple of intersected shapes.
 */
function intersection(shapes, options) {
    return new IntersectionShape(shapes, options);
}

/**
 * Creates the sync lazy-loaded shape.
 *
 * @param provider The provider that returns the base shape.
 */
function lazy(provider) {
    return new LazyShape(provider, false);
}
/**
 * Creates the sync lazy-loaded shape that only supports async parsing.
 *
 * @param provider The provider that returns the base shape.
 */
function lazyAsync(provider) {
    return new LazyShape(provider, true);
}

/**
 * Creates the integer shape that rejects `Infinity` and `NaN` values.
 *
 * @param options The constraint options or an issue message.
 */
function integer(options) {
    return new NumberShape(options).integer(options);
}

/**
 * Creates the constant value shape.
 *
 * @param value The value to which the input must be strictly equal.
 * @param options The constraint options or an issue message.
 * @template T The value type.
 */
function const_(value, options) {
    return new EnumShape([value], options);
}

/**
 * Creates the shape that always raises an issue.
 *
 * @param options The constraint options or an issue message.
 */
function never(options) {
    return new NeverShape(options);
}

/**
 * Creates the finite number shape that doesn't allow `Infinity` and `NaN` values.
 *
 * @param options The constraint options or an issue message.
 */
function number(options) {
    return new NumberShape(options);
}

/**
 * Creates the object shape.
 *
 * @param shapes The mapping from an object key to a corresponding shape.
 * @param options The constraint options or an issue message.
 */
function object(shapes, options) {
    return new ObjectShape(shapes, null, options);
}

/**
 * Synchronously preprocesses the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
function preprocess(cb) {
    return new Shape().transform(cb);
}
/**
 * Asynchronously preprocesses the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
function preprocessAsync(cb) {
    return new Shape().transformAsync(cb);
}

/**
 * Creates the shape of a value wrapped in a `Promise` instance.
 *
 * @param shape The shape of the resolved value.
 * @param options The constraint options or an issue message.
 * @template S The shape of the resolved value.
 */
function promise(shape, options) {
    return new PromiseShape(shape, options);
}

function record(keyShape, valueShape, options) {
    if (valueShape instanceof Shape) {
        return new RecordShape(keyShape, valueShape, options);
    }
    else {
        return new RecordShape(null, keyShape, options);
    }
}

/**
 * Creates the string shape.
 *
 * @param options The constraint options or an issue message.
 */
function string(options) {
    return new StringShape(options);
}

function tuple(shapes, restShape, options) {
    if (restShape == null || restShape instanceof Shape) {
        return new ArrayShape(shapes, restShape || null, options);
    }
    else {
        return new ArrayShape(shapes, null, restShape);
    }
}

/**
 * Creates a union shape that tries to parse the input with one of the provided shapes.
 *
 * @param shapes The list of shapes to try.
 * @param options The constraint options or an issue message.
 * @template U The tuple of united shapes.
 */
function union(shapes, options) {
    return new UnionShape(shapes, options);
}

/**
 * Creates the unconstrained shape with unknown value.
 */
function unknown() {
    return new Shape();
}

exports.ArrayShape = ArrayShape;
exports.BigIntShape = BigIntShape;
exports.BooleanShape = BooleanShape;
exports.EnumShape = EnumShape;
exports.ExcludeShape = ExcludeShape;
exports.InstanceShape = InstanceShape;
exports.IntersectionShape = IntersectionShape;
exports.LazyShape = LazyShape;
exports.NeverShape = NeverShape;
exports.NumberShape = NumberShape;
exports.ObjectShape = ObjectShape;
exports.PromiseShape = PromiseShape;
exports.RecordShape = RecordShape;
exports.RedirectShape = RedirectShape;
exports.ReplaceShape = ReplaceShape;
exports.Shape = Shape;
exports.StringShape = StringShape;
exports.TransformShape = TransformShape;
exports.UnionShape = UnionShape;
exports.ValidationError = ValidationError;
exports.and = intersection;
exports.any = any;
exports.array = array;
exports.bigint = bigint;
exports.boolean = boolean;
exports.const = const_;
exports.createIssueFactory = createIssueFactory;
exports.enum = enum_;
exports.instanceOf = instanceOf;
exports.int = integer;
exports.integer = integer;
exports.intersection = intersection;
exports.lazy = lazy;
exports.lazyAsync = lazyAsync;
exports.never = never;
exports.number = number;
exports.object = object;
exports.or = union;
exports.preprocess = preprocess;
exports.preprocessAsync = preprocessAsync;
exports.promise = promise;
exports.record = record;
exports.string = string;
exports.tuple = tuple;
exports.union = union;
exports.unknown = unknown;
