import { AnyShape, Shape } from './Shape';
import { ApplyResult, ConstraintOptions, Issue, Message, ParseOptions, TypeConstraintOptions } from '../shared-types';
export declare type InferTuple<U extends readonly AnyShape[], C extends 'input' | 'output'> = ToArray<{
    [K in keyof U]: U[K] extends AnyShape ? U[K][C] : never;
}>;
export declare type InferArray<U extends readonly AnyShape[] | null, R extends AnyShape | null, C extends 'input' | 'output'> = U extends readonly AnyShape[] ? R extends AnyShape ? [...InferTuple<U, C>, ...R[C][]] : InferTuple<U, C> : R extends AnyShape ? R[C][] : any[];
export declare type ToArray<T> = T extends any[] ? T : never;
/**
 * The shape of an array or a tuple.
 *
 * @template U The list of positioned element shapes or `null` if there are no positioned elements.
 * @template R The shape of rest elements or `null` if there are no rest elements.
 */
export declare class ArrayShape<U extends readonly AnyShape[] | null, R extends AnyShape | null> extends Shape<InferArray<U, R, 'input'>, InferArray<U, R, 'output'>> {
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    readonly shapes: U;
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    readonly restShape: R;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _issueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue;
    /**
     * Creates a new {@linkcode ArrayShape} instance.
     *
     * @param shapes The list of positioned element shapes or `null` if there are no positioned elements.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The type constraint options or the type issue message.
     */
    constructor(
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    shapes: U, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape: R, options?: TypeConstraintOptions | Message);
    at(key: unknown): AnyShape | null;
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template T The shape of rest elements.
     */
    rest<T extends AnyShape | null>(restShape: T): ArrayShape<U, T>;
    /**
     * Constrains the array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    length(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    min(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    max(length: number, options?: ConstraintOptions | Message): this;
    apply(input: unknown, options: ParseOptions): ApplyResult<InferArray<U, R, 'output'>>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<InferArray<U, R, 'output'>>>;
}
