import { __extends } from "tslib";
import { Shape } from './Shape';
import { appendCheck, arrayTypes, concatIssues, createIssueFactory, isArray, isAsyncShapes, isEqual, ok, unshiftPath, } from '../utils';
import { CODE_ARRAY_MAX, CODE_ARRAY_MIN, CODE_TUPLE, CODE_TYPE, MESSAGE_ARRAY_MAX, MESSAGE_ARRAY_MIN, MESSAGE_ARRAY_TYPE, MESSAGE_TUPLE, TYPE_ARRAY, } from '../constants';
var integerRegex = /^(?:0|[1-9]\d*)$/;
/**
 * The shape of an array or a tuple.
 *
 * @template U The list of positioned element shapes or `null` if there are no positioned elements.
 * @template R The shape of rest elements or `null` if there are no rest elements.
 */
var ArrayShape = /** @class */ (function (_super) {
    __extends(ArrayShape, _super);
    /**
     * Creates a new {@linkcode ArrayShape} instance.
     *
     * @param shapes The list of positioned element shapes or `null` if there are no positioned elements.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The type constraint options or the type issue message.
     */
    function ArrayShape(
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    shapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape, options) {
        var _this = _super.call(this, arrayTypes, (shapes !== null && isAsyncShapes(shapes)) || (restShape !== null && restShape.async)) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._options = options;
        if (shapes !== null && restShape === null) {
            _this._issueFactory = createIssueFactory(CODE_TUPLE, MESSAGE_TUPLE, options, shapes.length);
        }
        else {
            _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_ARRAY_TYPE, options, TYPE_ARRAY);
        }
        return _this;
    }
    ArrayShape.prototype.at = function (key) {
        var shapes = this.shapes;
        var index = typeof key === 'number' ? key : typeof key !== 'string' ? -1 : integerRegex.test(key) ? parseInt(key, 10) : -1;
        if (index % 1 !== 0 || index < 0) {
            return null;
        }
        if (shapes !== null && index < shapes.length) {
            return shapes[index];
        }
        return this.restShape;
    };
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template T The shape of rest elements.
     */
    ArrayShape.prototype.rest = function (restShape) {
        return new ArrayShape(this.shapes, restShape, this._options);
    };
    /**
     * Constrains the array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MIN, MESSAGE_ARRAY_MIN, options, length);
        return appendCheck(this, CODE_ARRAY_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MAX, MESSAGE_ARRAY_MAX, options, length);
        return appendCheck(this, CODE_ARRAY_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    ArrayShape.prototype.apply = function (input, options) {
        var _a = this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var inputLength;
        var shapesLength = 0;
        // noinspection CommaExpressionJS
        if (!isArray(input) ||
            ((inputLength = input.length),
                shapes !== null && inputLength !== (shapesLength = shapes.length) && restShape === null)) {
            return [this._issueFactory(input, options)];
        }
        var issues = null;
        var output = input;
        if (shapes !== null || restShape !== null) {
            for (var i = 0; i < inputLength; ++i) {
                var value = input[i];
                var valueShape = i < shapesLength ? shapes[i] : restShape;
                var result = valueShape.apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, i);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = input.slice(0);
                    }
                    output[i] = result.value;
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    ArrayShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            var _a = _this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var inputLength;
            var shapesLength = 0;
            // noinspection CommaExpressionJS
            if (!isArray(input) ||
                ((inputLength = input.length),
                    shapes !== null && inputLength !== (shapesLength = shapes.length) && restShape === null)) {
                resolve([_this._issueFactory(input, options)]);
                return;
            }
            var promises = [];
            if (shapes !== null || restShape !== null) {
                for (var i = 0; i < inputLength; ++i) {
                    var value = input[i];
                    var valueShape = i < shapesLength ? shapes[i] : restShape;
                    promises.push(valueShape.applyAsync(value, options));
                }
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var issues = null;
                var output = input;
                for (var i = 0; i < resultsLength; ++i) {
                    var result = results[i];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, i);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[i], result.value)) {
                        if (input === output) {
                            output = input.slice(0);
                        }
                        output[i] = result.value;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    return ArrayShape;
}(Shape));
export { ArrayShape };
