import { __extends } from "tslib";
import { Shape } from './Shape';
import { createIssueFactory, isArray, isAsyncShapes, isEqual, ok } from '../utils';
import { CODE_INTERSECTION, MESSAGE_INTERSECTION } from '../constants';
var IntersectionShape = /** @class */ (function (_super) {
    __extends(IntersectionShape, _super);
    function IntersectionShape(shapes, options) {
        var _this = _super.call(this, [], isAsyncShapes(shapes)) || this;
        _this.shapes = shapes;
        _this._issueFactory = createIssueFactory(CODE_INTERSECTION, MESSAGE_INTERSECTION, options, undefined);
        return _this;
    }
    IntersectionShape.prototype.apply = function (input, options) {
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var outputs = null;
        for (var i = 0; i < shapesLength; ++i) {
            var result = shapes[i].apply(input, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                return result;
            }
            var output = result.value;
            if (outputs === null) {
                outputs = [output];
                continue;
            }
            if (outputs.includes(output)) {
                continue;
            }
            outputs.push(output);
        }
        if (outputs === null) {
            return null;
        }
        if (outputs.length !== shapesLength) {
            outputs.push(input);
        }
        return intersectOutputs(input, outputs, this._issueFactory, options);
    };
    IntersectionShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var promises = [];
        for (var i = 0; i < shapesLength; ++i) {
            promises.push(shapes[i].applyAsync(input, options));
        }
        return Promise.all(promises).then(function (results) {
            var outputs = null;
            for (var i = 0; i < shapesLength; ++i) {
                var result = results[i];
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    return result;
                }
                var output = result.value;
                if (outputs === null) {
                    outputs = [output];
                    continue;
                }
                if (outputs.includes(output)) {
                    continue;
                }
                outputs.push(output);
            }
            if (outputs === null) {
                return null;
            }
            if (outputs.length !== shapesLength) {
                outputs.push(input);
            }
            return intersectOutputs(input, outputs, _this._issueFactory, options);
        });
    };
    return IntersectionShape;
}(Shape));
export { IntersectionShape };
var NEVER = Symbol();
function intersectOutputs(input, outputs, issueFactory, options) {
    var outputsLength = outputs.length;
    var value = outputs[0];
    for (var i = 1; i < outputsLength; ++i) {
        value = intersectPair(value, outputs[i]);
        if (value === NEVER) {
            return [issueFactory(input, options)];
        }
    }
    return ok(value);
}
function intersectPair(a, b) {
    if (isEqual(a, b)) {
        return a;
    }
    var aType = Shape.typeof(a);
    var bType = Shape.typeof(b);
    if (aType !== bType) {
        return NEVER;
    }
    if (aType === 'object') {
        var output = Object.assign({}, a, b);
        for (var key in a) {
            if (key in b) {
                var outputValue = intersectPair(a[key], b[key]);
                if (outputValue === NEVER) {
                    return NEVER;
                }
                output[key] = outputValue;
            }
        }
        return output;
    }
    if (aType === 'array') {
        var aLength = a.length;
        if (aLength === b.length) {
            var output = a;
            for (var i = 0; i < aLength; ++i) {
                var aValue = a[i];
                var bValue = b[i];
                if (isEqual(aValue, bValue)) {
                    continue;
                }
                if (output === a) {
                    output = a.slice(0);
                }
                var outputValue = intersectPair(aValue, bValue);
                if (outputValue === NEVER) {
                    return NEVER;
                }
                output[i] = outputValue;
            }
            return output;
        }
    }
    return NEVER;
}
