import { Shape } from './Shape';
import { ApplyResult, ConstraintOptions, Message, ParseOptions, TypeConstraintOptions } from '../shared-types';
/**
 * The shape of the finite number.
 */
export declare class NumberShape extends Shape<number> {
    protected _issueFactory: (input: unknown, options: Readonly<ParseOptions>) => import("../shared-types").Issue;
    protected _typePredicate: (number: unknown) => boolean;
    /**
     * Creates a new {@linkcode NumberShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    /**
     * Constrains the number to be greater than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    positive(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    negative(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be greater than the value.
     *
     * @param value The exclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    gt(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than the value.
     *
     * @param value The exclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    lt(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    gte(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    lte(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param value The number by which the input should be divisible without a remainder.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    multipleOf(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be an integer and rejects `Infinity` and `NaN` values.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    integer(options?: ConstraintOptions | Message): this;
    apply(input: unknown, options: ParseOptions): ApplyResult<number>;
}
