import { __extends } from "tslib";
import { Shape } from './Shape';
import { appendCheck, createIssueFactory, numberTypes } from '../utils';
import { CODE_NUMBER_GT, CODE_NUMBER_GTE, CODE_NUMBER_LT, CODE_NUMBER_LTE, CODE_NUMBER_MULTIPLE_OF, CODE_TYPE, MESSAGE_INTEGER_TYPE, MESSAGE_NUMBER_GT, MESSAGE_NUMBER_GTE, MESSAGE_NUMBER_LT, MESSAGE_NUMBER_LTE, MESSAGE_NUMBER_MULTIPLE_OF, MESSAGE_NUMBER_TYPE, TYPE_INTEGER, TYPE_NUMBER, } from '../constants';
/**
 * The shape of the finite number.
 */
var NumberShape = /** @class */ (function (_super) {
    __extends(NumberShape, _super);
    /**
     * Creates a new {@linkcode NumberShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function NumberShape(options) {
        var _this = _super.call(this, numberTypes) || this;
        _this._typePredicate = Number.isFinite;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_NUMBER_TYPE, options, TYPE_NUMBER);
        return _this;
    }
    /**
     * Constrains the number to be greater than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.positive = function (options) {
        return this.gt(0, options);
    };
    /**
     * Constrains the number to be less than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.negative = function (options) {
        return this.lt(0, options);
    };
    /**
     * Constrains the number to be greater than the value.
     *
     * @param value The exclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GT, MESSAGE_NUMBER_GT, options, value);
        return appendCheck(this, CODE_NUMBER_GT, options, value, function (input, options) {
            if (input <= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than the value.
     *
     * @param value The exclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LT, MESSAGE_NUMBER_LT, options, value);
        return appendCheck(this, CODE_NUMBER_LT, options, value, function (input, options) {
            if (input >= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GTE, MESSAGE_NUMBER_GTE, options, value);
        return appendCheck(this, CODE_NUMBER_GTE, options, value, function (input, options) {
            if (input < value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LTE, MESSAGE_NUMBER_LTE, options, value);
        return appendCheck(this, CODE_NUMBER_LTE, options, value, function (input, options) {
            if (input > value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param value The number by which the input should be divisible without a remainder.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.multipleOf = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_MULTIPLE_OF, MESSAGE_NUMBER_MULTIPLE_OF, options, value);
        return appendCheck(this, CODE_NUMBER_MULTIPLE_OF, options, value, function (input, options) {
            if (input % value !== 0) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be an integer and rejects `Infinity` and `NaN` values.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.integer = function (options) {
        var shape = this._clone();
        shape._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_INTEGER_TYPE, options, TYPE_INTEGER);
        shape._typePredicate = Number.isInteger;
        return shape;
    };
    NumberShape.prototype.apply = function (input, options) {
        var _a = this, _typePredicate = _a._typePredicate, _applyChecks = _a._applyChecks;
        if (!_typePredicate(input)) {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return NumberShape;
}(Shape));
export { NumberShape };
