import { ApplyResult, Issue, Message, ParseOptions, ReadonlyDict, TypeConstraintOptions } from '../shared-types';
import { isObjectLike } from '../utils';
import { AnyShape, OpaqueExclude, OpaqueReplace, Shape } from './Shape';
import { EnumShape } from './EnumShape';
export declare type InferObject<P extends ReadonlyDict<AnyShape>, R extends AnyShape | null, C extends 'input' | 'output'> = Squash<UndefinedAsOptional<{
    [K in keyof P]: P[K][C];
}> & InferIndexer<R, C>>;
export declare type InferIndexer<R extends AnyShape | null, C extends 'input' | 'output'> = R extends Shape ? {
    [key: string]: R[C];
} : unknown;
export declare type StringKeyof<T extends object> = Extract<keyof T, string>;
export declare type Squash<T> = T extends never ? never : {
    [K in keyof T]: T[K];
};
export declare type UndefinedAsOptional<T> = OmitBy<T, undefined> & Partial<PickBy<T, undefined>>;
export declare type OmitBy<T, V> = Omit<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
export declare type PickBy<T, V> = Pick<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
export declare type Optional<P extends ReadonlyDict<AnyShape>> = {
    [K in keyof P]: OpaqueReplace<P[K], undefined>;
};
export declare type Required<P extends ReadonlyDict<AnyShape>> = {
    [K in keyof P]: OpaqueExclude<P[K], undefined>;
};
export declare type KeysMode = 'preserved' | 'stripped' | 'exact';
/**
 * The shape of an object.
 *
 * @template P The mapping from an object key to a corresponding value shape.
 * @template R The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
 */
export declare class ObjectShape<P extends ReadonlyDict<AnyShape>, R extends AnyShape | null> extends Shape<InferObject<P, R, 'input'>, InferObject<P, R, 'output'>> {
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    readonly shapes: P;
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    readonly restShape: R;
    /**
     * The array of known object keys.
     */
    readonly keys: readonly StringKeyof<P>[];
    /**
     * The mode of unknown keys handling.
     */
    readonly keysMode: KeysMode;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _valueShapes: Shape[];
    protected _typePredicate: typeof isObjectLike;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue;
    protected _exactIssueFactory?: (input: unknown, options: Readonly<ParseOptions>, param: unknown) => Issue;
    /**
     * Creates a new {@linkcode ObjectShape} instance.
     *
     * @param shapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The type constraint options or an issue message.
     * @param keysMode The mode of unknown keys handling.
     * @template P The mapping from an object key to a corresponding value shape.
     * @template R The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    constructor(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    shapes: P, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape: R, options?: TypeConstraintOptions | Message, keysMode?: KeysMode);
    at(key: any): AnyShape | null;
    /**
     * Merge properties from the other object shape.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@linkcode keysMode} is preserved intact.
     *
     * The returned object shape would have no checks.
     *
     * @param shape The object shape which properties must be added to this object shape.
     * @returns The new object shape.
     * @template T Properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shape: ObjectShape<T, any>): ObjectShape<Pick<P, Exclude<keyof P, keyof T>> & T, R>;
    /**
     * Add properties to an object shape.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@linkcode keysMode} is preserved intact.
     *
     * The returned object shape would have no checks.
     *
     * @param shapes The properties to add.
     * @returns The new object shape.
     * @template T The shapes of properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shapes: T): ObjectShape<Pick<P, Exclude<keyof P, keyof T>> & T, R>;
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    pick<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Pick<P, K[number]>, R>;
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    omit<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]>, R>;
    /**
     * Returns an object shape with all properties marked as optional.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    partial(): ObjectShape<Optional<P>, R>;
    /**
     * Returns an object shape with keys marked as optional.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to make optional.
     * @returns The new object shape.
     * @template K The list of string keys.
     */
    partial<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]> & Optional<Pick<P, K[number]>>, R>;
    /**
     * Returns an object shape with all properties marked as required.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    required(): ObjectShape<Required<P>, R>;
    /**
     * Returns an object shape with keys marked as required.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to make required.
     * @returns The new object shape.
     * @template K The list of string keys.
     */
    required<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]> & Required<Pick<P, K[number]>>, R>;
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * The returned object shape would have no checks.
     *
     * @param options The constraint options or an issue message.
     * @returns The new object shape.
     */
    exact(options?: TypeConstraintOptions | Message): ObjectShape<P, null>;
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    strip(): ObjectShape<P, null>;
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    preserve(): ObjectShape<P, null>;
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape that validates values at unknown keys.
     * @returns The new object shape.
     * @template T The index signature shape.
     */
    rest<T extends AnyShape>(restShape: T): ObjectShape<P, T>;
    /**
     * Returns the enum shape of keys of this object.
     */
    keyof(): EnumShape<StringKeyof<P>>;
    /**
     * Constrains an object to be an `Object` instance or to have a `null` prototype.
     */
    plain(): this;
    apply(input: unknown, options: ParseOptions): ApplyResult<InferObject<P, R, 'output'>>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<InferObject<P, R, 'output'>>>;
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    private _applyLax;
    /**
     * Unknown keys are either parsed with a {@linkcode restShape}, stripped, or cause an issue.
     */
    private _applyStrict;
}
