import { __extends } from "tslib";
import { CODE_TYPE, CODE_UNKNOWN_KEYS, MESSAGE_OBJECT_TYPE, MESSAGE_UNKNOWN_KEYS, TYPE_OBJECT } from '../constants';
import { cloneEnumerableKeys, cloneKnownKeys, concatIssues, createIssueFactory, isArray, isAsyncShapes, isEqual, isFlagSet, isObjectLike, isPlainObject, objectTypes, ok, pushIssue, setFlag, setKeyValue, unshiftPath, } from '../utils';
import { Shape } from './Shape';
import { EnumShape } from './EnumShape';
/**
 * The shape of an object.
 *
 * @template P The mapping from an object key to a corresponding value shape.
 * @template R The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
 */
var ObjectShape = /** @class */ (function (_super) {
    __extends(ObjectShape, _super);
    /**
     * Creates a new {@linkcode ObjectShape} instance.
     *
     * @param shapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The type constraint options or an issue message.
     * @param keysMode The mode of unknown keys handling.
     * @template P The mapping from an object key to a corresponding value shape.
     * @template R The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    function ObjectShape(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    shapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape, options, keysMode) {
        if (keysMode === void 0) { keysMode = 'preserved'; }
        var _this = this;
        var valueShapes = Object.values(shapes);
        _this = _super.call(this, objectTypes, isAsyncShapes(valueShapes) || (restShape !== null && restShape.async)) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._typePredicate = isObjectLike;
        _this.keys = Object.keys(shapes);
        _this.keysMode = keysMode;
        _this._options = options;
        _this._valueShapes = valueShapes;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    ObjectShape.prototype.at = function (key) {
        return this.shapes.hasOwnProperty(key) ? this.shapes[key] : this.restShape;
    };
    ObjectShape.prototype.extend = function (shape) {
        var shapes = Object.assign({}, this.shapes, shape instanceof ObjectShape ? shape.shapes : shape);
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    ObjectShape.prototype.pick = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    ObjectShape.prototype.omit = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (!keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.partial = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.optional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.required = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.nonOptional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * The returned object shape would have no checks.
     *
     * @param options The constraint options or an issue message.
     * @returns The new object shape.
     */
    ObjectShape.prototype.exact = function (options) {
        var shape = new ObjectShape(this.shapes, null, this._options, 'exact');
        shape._exactIssueFactory = createIssueFactory(CODE_UNKNOWN_KEYS, MESSAGE_UNKNOWN_KEYS, options);
        return shape;
    };
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.strip = function () {
        return new ObjectShape(this.shapes, null, this._options, 'stripped');
    };
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.preserve = function () {
        return new ObjectShape(this.shapes, null, this._options);
    };
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape that validates values at unknown keys.
     * @returns The new object shape.
     * @template T The index signature shape.
     */
    ObjectShape.prototype.rest = function (restShape) {
        return new ObjectShape(this.shapes, restShape, this._options);
    };
    /**
     * Returns the enum shape of keys of this object.
     */
    ObjectShape.prototype.keyof = function () {
        return new EnumShape(this.keys);
    };
    /**
     * Constrains an object to be an `Object` instance or to have a `null` prototype.
     */
    ObjectShape.prototype.plain = function () {
        var shape = this._clone();
        shape._typePredicate = isPlainObject;
        return shape;
    };
    ObjectShape.prototype.apply = function (input, options) {
        var _typePredicate = this._typePredicate;
        if (!_typePredicate(input)) {
            return [this._typeIssueFactory(input, options)];
        }
        if (this.keysMode === 'preserved' && this.restShape === null) {
            return this._applyLax(input, options);
        }
        else {
            return this._applyStrict(input, options);
        }
    };
    ObjectShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            var _typePredicate = _this._typePredicate;
            if (!_typePredicate(input)) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var _a = _this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var keysLength = keys.length;
            var promises = [];
            var issues = null;
            var output = input;
            var seenCount = 0;
            var seenFlags = 0;
            var unknownKeys = null;
            for (var key in input) {
                var value = input[key];
                var index = keys.indexOf(key);
                var valueShape = restShape;
                if (index !== -1) {
                    seenCount++;
                    seenFlags = setFlag(seenFlags, index);
                    valueShape = _valueShapes[index];
                }
                if (valueShape !== null) {
                    promises.push(key, valueShape.applyAsync(value, options));
                    continue;
                }
                if (keysMode === 'exact') {
                    if (unknownKeys !== null) {
                        unknownKeys.push(key);
                        continue;
                    }
                    unknownKeys = [key];
                    if (!options.verbose) {
                        break;
                    }
                    continue;
                }
                if (input === output && keysMode === 'stripped') {
                    output = cloneKnownKeys(input, keys);
                }
            }
            if (unknownKeys !== null) {
                var issue = _this._exactIssueFactory(input, options, unknownKeys);
                if (!options.verbose) {
                    resolve([issue]);
                    return;
                }
                issues = pushIssue(issues, issue);
            }
            if (seenCount !== keysLength) {
                for (var i = 0; i < keysLength; ++i) {
                    if (isFlagSet(seenFlags, i)) {
                        continue;
                    }
                    var key = keys[i];
                    var value = input[key];
                    promises.push(key, _valueShapes[i].applyAsync(value, options));
                }
            }
            resolve(Promise.all(promises).then(function (entries) {
                var entriesLength = entries.length;
                for (var i = 0; i < entriesLength; i += 2) {
                    var key = entries[i];
                    var result = entries[i + 1];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, key);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[key], result.value)) {
                        if (input === output) {
                            output = cloneEnumerableKeys(input);
                        }
                        setKeyValue(output, key, result.value);
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    ObjectShape.prototype._applyLax = function (input, options) {
        var _a = this, keys = _a.keys, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        for (var i = 0; i < keysLength; ++i) {
            var key = keys[i];
            var value = input[key];
            var result = _valueShapes[i].apply(value, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftPath(result, key);
                if (!options.verbose) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                if (input === output) {
                    output = cloneEnumerableKeys(input);
                }
                setKeyValue(output, key, result.value);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Unknown keys are either parsed with a {@linkcode restShape}, stripped, or cause an issue.
     */
    ObjectShape.prototype._applyStrict = function (input, options) {
        var _a = this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        var seenCount = 0;
        var seenFlags = 0;
        var unknownKeys = null;
        for (var key in input) {
            var value = input[key];
            var index = keys.indexOf(key);
            var valueShape = restShape;
            // The key is known
            if (index !== -1) {
                seenCount++;
                seenFlags = setFlag(seenFlags, index);
                valueShape = _valueShapes[index];
            }
            // The key is known or indexed
            if (valueShape !== null) {
                var result = valueShape.apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = restShape === null ? cloneKnownKeys(input, keys) : cloneEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
                continue;
            }
            // Unknown keys raise an issue
            if (keysMode === 'exact') {
                if (unknownKeys !== null) {
                    unknownKeys.push(key);
                    continue;
                }
                unknownKeys = [key];
                if (!options.verbose) {
                    break;
                }
                continue;
            }
            // Unknown keys are stripped
            if (input === output && (_unsafe || issues === null)) {
                output = cloneKnownKeys(input, keys);
            }
        }
        // Raise unknown keys issue
        if (unknownKeys !== null) {
            var issue = this._exactIssueFactory(input, options, unknownKeys);
            if (!options.verbose) {
                return [issue];
            }
            issues = pushIssue(issues, issue);
        }
        // Parse absent known keys
        if (seenCount !== keysLength) {
            for (var i = 0; i < keysLength; ++i) {
                if (isFlagSet(seenFlags, i)) {
                    continue;
                }
                var key = keys[i];
                var value = input[key];
                var result = _valueShapes[i].apply(value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = cloneEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    return ObjectShape;
}(Shape));
export { ObjectShape };
