import { __extends } from "tslib";
import { Shape } from './Shape';
import { cloneEnumerableKeys, concatIssues, createIssueFactory, isArray, isEqual, isObjectLike, objectTypes, ok, setKeyValue, unshiftPath, } from '../utils';
import { CODE_TYPE, MESSAGE_OBJECT_TYPE, TYPE_OBJECT } from '../constants';
/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
var RecordShape = /** @class */ (function (_super) {
    __extends(RecordShape, _super);
    /**
     * Creates a new {@linkcode RecordShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape or `null` if keys should be preserved intact.
     * @template V The value shape.
     */
    function RecordShape(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this, objectTypes, (keyShape != null && keyShape.async) || valueShape.async) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    RecordShape.prototype.at = function (key) {
        return typeof key === 'string' || typeof key === 'number' ? this.valueShape : null;
    };
    RecordShape.prototype.apply = function (input, options) {
        if (!isObjectLike(input)) {
            return [this._issueFactory(input, options)];
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keyCount = 0;
        var issues = null;
        var output = input;
        for (var key in input) {
            var value = input[key];
            var outputKey = key;
            var outputValue = value;
            if (keyShape !== null) {
                var keyResult = keyShape.apply(key, options);
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftPath(keyResult, key);
                        if (!options.verbose) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                    }
                    else {
                        outputKey = keyResult.value;
                    }
                }
            }
            var valueResult = valueShape.apply(value, options);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftPath(valueResult, key);
                    if (!options.verbose) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                }
                else {
                    outputValue = valueResult.value;
                }
            }
            if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                if (input === output) {
                    output = cloneEnumerableKeys(input, keyCount);
                }
                setKeyValue(output, outputKey, outputValue);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    RecordShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        return new Promise(function (resolve) {
            if (!isObjectLike(input)) {
                resolve([_this._issueFactory(input, options)]);
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var promises = [];
            for (var key in input) {
                var value = input[key];
                promises.push(key, keyShape !== null ? keyShape.applyAsync(key, options) : null, valueShape.applyAsync(value, options));
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var keyCount = 0;
                var issues = null;
                var output = input;
                for (var i = 0; i < resultsLength; i += 3) {
                    var key = results[i];
                    var value = input[key];
                    var outputKey = key;
                    var outputValue = value;
                    var keyResult = results[i + 1];
                    var valueResult = results[i + 2];
                    if (keyResult !== null) {
                        if (isArray(keyResult)) {
                            unshiftPath(keyResult, key);
                            if (!options.verbose) {
                                return keyResult;
                            }
                            issues = concatIssues(issues, keyResult);
                        }
                        else {
                            outputKey = keyResult.value;
                        }
                    }
                    if (valueResult !== null) {
                        if (isArray(valueResult)) {
                            unshiftPath(valueResult, key);
                            if (!options.verbose) {
                                return valueResult;
                            }
                            issues = concatIssues(issues, valueResult);
                        }
                        else {
                            outputValue = valueResult.value;
                        }
                    }
                    if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                        if (input === output) {
                            output = cloneEnumerableKeys(input, keyCount);
                        }
                        output[outputKey] = outputValue;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    return RecordShape;
}(Shape));
export { RecordShape };
