import { ApplyChecksCallback, ApplyResult, Check, CheckCallback, CheckOptions, Err, Message, Ok, ParseOptions, RefineOptions, TypeConstraintOptions } from '../shared-types';
/**
 * An arbitrary shape.
 */
export declare type AnyShape = Shape | Shape<never>;
/**
 * A shape that replaces a value from an input with an output value.
 *
 * @template S The base shape.
 * @template A The searched value.
 * @template B The value that is used as a replacement.
 */
export declare type OpaqueReplace<S extends AnyShape, A, B = A> = Shape<S['input'] | A, Exclude<S['output'], A> | B>;
/**
 * Excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
export declare type OpaqueExclude<S extends AnyShape, T> = Shape<Exclude<S['input'], T>, Exclude<S['output'], T>>;
/**
 * The detected runtime input value type.
 */
export declare type ValueType = 'object' | 'array' | 'function' | 'string' | 'symbol' | 'number' | 'bigint' | 'boolean' | 'null' | 'undefined' | 'any' | 'never';
export interface Shape<I, O> {
    /**
     * The shape input type. Accessible only at compile time for type inference.
     */
    readonly input: I;
    /**
     * The shape output type. Accessible only at compile time for type inference.
     */
    readonly output: O;
    /**
     * Synchronously parses the value and returns {@linkcode Ok} or {@linkcode Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@linkcode Ok} instance if parsing has succeeded or {@linkcode Err} if parsing has failed.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     */
    try(input: unknown, options?: ParseOptions): Ok<O> | Err;
    /**
     * Asynchronously parses the value and returns {@linkcode Ok} or {@linkcode Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@linkcode Ok} instance if parsing has succeeded or {@linkcode Err} if parsing has failed.
     */
    tryAsync(input: unknown, options?: ParseOptions): Promise<Ok<O> | Err>;
    /**
     * Synchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     * @throws {@linkcode ValidationError} if any issues occur during parsing.
     */
    parse(input: unknown, options?: ParseOptions): O;
    /**
     * Asynchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws {@linkcode ValidationError} if any issues occur during parsing.
     */
    parseAsync(input: unknown, options?: ParseOptions): Promise<O>;
    /**
     * Synchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     */
    parseOrDefault<T = undefined>(input: unknown, defaultValue?: T, options?: ParseOptions): O | T;
    /**
     * Asynchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync<T = undefined>(input: unknown, defaultValue?: T, options?: ParseOptions): Promise<O | T>;
}
/**
 * The base shape.
 *
 * @template I The input value.
 * @template O The output value.
 */
export declare class Shape<I = any, O = I> {
    /**
     * `true` if the shape allows only {@linkcode parseAsync} and throws an error if {@linkcode parse} is called.
     * `false` if the shape can be used in both sync and async contexts.
     */
    readonly async: boolean;
    /**
     * The list of runtime value types that can be processed by the shape. Use {@linkcode Shape.typeof} to detect a
     * runtime value type.
     */
    readonly inputTypes: readonly ValueType[];
    /**
     * The list of checks applied to the shape output.
     */
    readonly checks: readonly Check[];
    /**
     * The human-readable shape description.
     */
    description: string;
    /**
     * Applies checks to the output.
     */
    protected _applyChecks: ApplyChecksCallback | null;
    /**
     * `true` if some checks from {@linkcode checks} were marked as unsafe, `false` otherwise.
     */
    protected _unsafe: boolean;
    /**
     * Creates the new {@linkcode Shape} instance.
     *
     * @param inputTypes
     * @param async If `true` then the shape would allow only {@linkcode parseAsync} and throw an error if
     * {@linkcode parse} is called. Otherwise, shape can be used in both sync and async contexts.
     * @template I The input value.
     * @template O The output value.
     */
    constructor(inputTypes?: readonly ValueType[], async?: boolean);
    /**
     * Returns the type of the value.
     *
     * @param value The value to detect type of.
     */
    static typeof(value: unknown): ValueType;
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    at(key: unknown): AnyShape | null;
    /**
     * Adds a human-readable description text to the shape.
     *
     * @param text The description text.
     * @returns The clone of this shape with the description added.
     */
    describe(text: string): this;
    /**
     * Appends the check that is applied to the shape output.
     *
     * If the {@linkcode CheckOptions.key} is defined and there's already a check with the same key then, it is removed
     * and a new check is appended to the list of shape checks. If the key is `undefined` then the `cb` identity is used
     * as a key.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The check options.
     * @returns The clone of this shape with the check added.
     */
    check(cb: CheckCallback<O>, options?: CheckOptions): this;
    /**
     * Redirects the shape output to another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @returns The {@linkcode RedirectShape} instance.
     * @template T The output value.
     */
    to<T>(shape: Shape<O, T>): Shape<I, T>;
    /**
     * Synchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The output value.
     */
    transform<T>(
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb: (output: O, options: Readonly<ParseOptions>) => T): Shape<I, T>;
    /**
     * Asynchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The transformed value.
     */
    transformAsync<T>(
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb: (output: O, options: Readonly<ParseOptions>) => Promise<T>): Shape<I, T>;
    /**
     * Refines the shape output type with the [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if value conforms the required type, or `false` otherwise.
     * @param options The constraint options or an issue message.
     * @returns The shape with the narrowed output.
     * @template T The narrowed output value.
     */
    refine<T extends O>(
    /**
     * @param output The shape output value.
     */
    cb: (output: O) => output is T, options?: RefineOptions | Message): Shape<I, T>;
    /**
     * Checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
     * @param options The constraint options or an issue message.
     * @returns The clone of this shape.
     */
    refine(
    /**
     * @param output The shape output value.
     */
    cb: (output: O) => boolean, options?: RefineOptions | Message): this;
    /**
     * Replaces `undefined` input value with an `undefined` output value.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    optional(): OpaqueReplace<this, undefined>;
    /**
     * Replaces `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    optional<T>(defaultValue: T): OpaqueReplace<this, undefined, T>;
    /**
     * Replaces `null` input value with an `null` output value.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullable(): OpaqueReplace<this, null>;
    /**
     * Replaces `null` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `null`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullable<T>(defaultValue: T): OpaqueReplace<this, null, T>;
    /**
     * Passes `null` and `undefined` input value to the output.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullish(): OpaqueReplace<this, null | undefined>;
    /**
     * Replaces `null` and `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined` or `null`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullish<T>(defaultValue?: T): OpaqueReplace<this, null | undefined, T>;
    /**
     * Prevents an input and output from being undefined.
     *
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     */
    nonOptional(options?: TypeConstraintOptions | Message): OpaqueExclude<this, undefined>;
    /**
     * Synchronously parses the input.
     *
     * Use {@linkcode parse} or {@linkcode try} instead of this method whenever possible. Override this method to
     * implement a custom shape.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    apply(input: unknown, options: ParseOptions): ApplyResult<O>;
    /**
     * Asynchronously parses the input.
     *
     * Use {@linkcode parseAsync} or {@linkcode tryAsync} instead of this method whenever possible. Override this method
     * to implement a custom shape that requires an async context.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O>>;
    /**
     * Returns the shallow clone of this shape.
     */
    protected _clone(): this;
}
/**
 * The shape that applies a transformer callback to the base shape output.
 *
 * @template S The base shape.
 * @template O The transformed value.
 */
export declare class TransformShape<S extends AnyShape, O> extends Shape<S['input'], O> {
    /**
     * The base shape which output value is transformed.
     */
    readonly shape: S;
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    readonly callback: (output: S['output'], options: Readonly<ParseOptions>) => Promise<O> | O;
    /**
     * Creates the new {@linkcode TransformShape} instance.
     *
     * @param shape The base shape.
     * @param async If `true` then the transformed shape would await for the transformer to finish and use the resolved
     * value as an output. Otherwise, the value that is synchronously returned from the transformer is used as an output.
     * @param callback The callback that transforms the shape output value.
     * @template S The base shape.
     * @template O The transformed value.
     */
    constructor(
    /**
     * The base shape which output value is transformed.
     */
    shape: S, async: boolean, 
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    callback: (output: S['output'], options: Readonly<ParseOptions>) => Promise<O> | O);
    apply(input: unknown, options: ParseOptions): ApplyResult<O>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O>>;
}
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template I The input shape.
 * @template O The output shape.
 */
export declare class RedirectShape<I extends AnyShape, O extends Shape<I['output'], any>> extends Shape<I['input'], O['output']> {
    /**
     * The input shape.
     */
    readonly inputShape: I;
    /**
     * The output shape.
     */
    readonly outputShape: O;
    /**
     * Creates the new {@linkcode RedirectShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template I The input shape.
     * @template O The output shape.
     */
    constructor(
    /**
     * The input shape.
     */
    inputShape: I, 
    /**
     * The output shape.
     */
    outputShape: O);
    apply(input: unknown, options: ParseOptions): ApplyResult<O['output']>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O['output']>>;
}
/**
 * The shape that replaces an input value with another value.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template A The replaced value.
 * @template B The replacement value.
 */
export declare class ReplaceShape<S extends AnyShape, A, B = A> extends Shape<S['input'] | A, Exclude<S['output'], A> | B> {
    /**
     * The shape that parses the input without the replaced value.
     */
    readonly shape: S;
    /**
     * The replaced value.
     */
    readonly searchedValue: A;
    /**
     * The replacement value.
     */
    readonly value?: B | undefined;
    private _replacedResult;
    private _replacedResultPromise?;
    /**
     * Creates the new {@linkcode ReplaceShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param searchedValue The input that should be replaced.
     * @param value The replacement value. If `undefined` then the replaced value is returned as is without any parsing.
     * @template S The shape that parses the input without the replaced value.
     * @template A The searched value.
     * @template B The replacement value.
     */
    constructor(
    /**
     * The shape that parses the input without the replaced value.
     */
    shape: S, 
    /**
     * The replaced value.
     */
    searchedValue: A, 
    /**
     * The replacement value.
     */
    value?: B | undefined);
    apply(input: unknown, options: ParseOptions): ApplyResult<Exclude<S['output'], A> | B>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<Exclude<S['output'], A> | B>>;
}
/**
 * The shape that excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
export declare class ExcludeShape<S extends AnyShape, T> extends Shape<Exclude<S['input'], T>, Exclude<S['output'], T>> {
    /**
     * The base shape.
     */
    readonly shape: S;
    /**
     * The excluded value.
     */
    readonly excludedValue: T;
    protected _issueFactory: (input: unknown, options: Readonly<ParseOptions>) => import("../shared-types").Issue;
    /**
     * Creates the new {@linkcode ExcludeShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param excludedValue The excluded value.
     * @param options The constraint options or an issue message.
     * @template S The shape that parses the input without the replaced value.
     * @template T The excluded value.
     */
    constructor(
    /**
     * The base shape.
     */
    shape: S, 
    /**
     * The excluded value.
     */
    excludedValue: T, options?: TypeConstraintOptions | Message);
    apply(input: unknown, options: ParseOptions): ApplyResult<Exclude<S['output'], T>>;
    applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<Exclude<S['output'], T>>>;
}
