import { __extends } from "tslib";
import { anyTypes, appendCheck, captureIssues, createApplyChecksCallback, createIssueFactory, getPrototypeOf, isArray, isEqual, ok, unique, } from '../utils';
import { ValidationError } from '../ValidationError';
import { CODE_EXCLUSION, CODE_PREDICATE, MESSAGE_ERROR_ASYNC, MESSAGE_EXCLUSION, MESSAGE_PREDICATE, } from '../constants';
var defaultParseOptions = Object.freeze({ verbose: false });
/**
 * The base shape.
 *
 * @template I The input value.
 * @template O The output value.
 */
var Shape = /** @class */ (function () {
    /**
     * Creates the new {@linkcode Shape} instance.
     *
     * @param inputTypes
     * @param async If `true` then the shape would allow only {@linkcode parseAsync} and throw an error if
     * {@linkcode parse} is called. Otherwise, shape can be used in both sync and async contexts.
     * @template I The input value.
     * @template O The output value.
     */
    function Shape(inputTypes, async) {
        if (async === void 0) { async = false; }
        /**
         * The list of checks applied to the shape output.
         */
        this.checks = [];
        /**
         * The human-readable shape description.
         */
        this.description = '';
        /**
         * Applies checks to the output.
         */
        this._applyChecks = null;
        /**
         * `true` if some checks from {@linkcode checks} were marked as unsafe, `false` otherwise.
         */
        this._unsafe = false;
        this.inputTypes =
            inputTypes === undefined || inputTypes.length === 0 || inputTypes.includes('any') ? anyTypes : unique(inputTypes);
        this.async = async;
        if (async) {
            this.apply = function () {
                throw new Error(MESSAGE_ERROR_ASYNC);
            };
        }
    }
    /**
     * Returns the type of the value.
     *
     * @param value The value to detect type of.
     */
    Shape.typeof = function (value) {
        var type = typeof value;
        if (type !== 'object') {
            return type;
        }
        if (value === null) {
            return 'null';
        }
        if (isArray(value)) {
            return 'array';
        }
        return type;
    };
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    Shape.prototype.at = function (key) {
        return null;
    };
    /**
     * Adds a human-readable description text to the shape.
     *
     * @param text The description text.
     * @returns The clone of this shape with the description added.
     */
    Shape.prototype.describe = function (text) {
        var shape = this._clone();
        shape.description = text;
        return shape;
    };
    /**
     * Appends the check that is applied to the shape output.
     *
     * If the {@linkcode CheckOptions.key} is defined and there's already a check with the same key then, it is removed
     * and a new check is appended to the list of shape checks. If the key is `undefined` then the `cb` identity is used
     * as a key.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The check options.
     * @returns The clone of this shape with the check added.
     */
    Shape.prototype.check = function (cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.key, key = _a === void 0 ? cb : _a, _b = options.unsafe, unsafe = _b === void 0 ? false : _b, param = options.param;
        var checks = this.checks.filter(function (check) { return check.key !== key; });
        checks.push({ key: key, callback: cb, unsafe: unsafe, param: param });
        var shape = this._clone();
        shape.checks = checks;
        shape._applyChecks = createApplyChecksCallback(checks);
        shape._unsafe || (shape._unsafe = unsafe);
        return shape;
    };
    /**
     * Redirects the shape output to another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @returns The {@linkcode RedirectShape} instance.
     * @template T The output value.
     */
    Shape.prototype.to = function (shape) {
        return new RedirectShape(this, shape);
    };
    /**
     * Synchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The output value.
     */
    Shape.prototype.transform = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, false, cb);
    };
    /**
     * Asynchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The transformed value.
     */
    Shape.prototype.transformAsync = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, true, cb);
    };
    Shape.prototype.refine = function (cb, options) {
        var issueFactory = createIssueFactory(CODE_PREDICATE, MESSAGE_PREDICATE, options, cb);
        return appendCheck(this, undefined, options, cb, function (input, options) {
            if (!cb(input)) {
                return issueFactory(input, options);
            }
        });
    };
    Shape.prototype.optional = function (defaultValue) {
        return new ReplaceShape(this, undefined, defaultValue);
    };
    Shape.prototype.nullable = function (defaultValue) {
        return new ReplaceShape(this, null, defaultValue);
    };
    Shape.prototype.nullish = function (defaultValue) {
        return this.nullable(defaultValue).optional(defaultValue);
    };
    /**
     * Prevents an input and output from being undefined.
     *
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     */
    Shape.prototype.nonOptional = function (options) {
        var shape = this;
        while (shape instanceof ReplaceShape && shape.checks.length === 0 && shape.searchedValue === undefined) {
            shape = shape.shape;
        }
        return new ExcludeShape(shape, undefined, options);
    };
    /**
     * Synchronously parses the input.
     *
     * Use {@linkcode parse} or {@linkcode try} instead of this method whenever possible. Override this method to
     * implement a custom shape.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    /**
     * Asynchronously parses the input.
     *
     * Use {@linkcode parseAsync} or {@linkcode tryAsync} instead of this method whenever possible. Override this method
     * to implement a custom shape that requires an async context.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype.applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) { return resolve(_this.apply(input, options)); });
    };
    /**
     * Returns the shallow clone of this shape.
     */
    Shape.prototype._clone = function () {
        return Object.assign(Object.create(getPrototypeOf(this)), this);
    };
    return Shape;
}());
export { Shape };
Object.defineProperty(Shape.prototype, 'try', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return ok(input);
            }
            if (isArray(result)) {
                return { ok: false, issues: result };
            }
            return ok(result.value);
        };
        Object.defineProperty(this, 'try', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'tryAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return ok(input);
                }
                if (isArray(result)) {
                    return { ok: false, issues: result };
                }
                return result;
            });
        };
        Object.defineProperty(this, 'tryAsync', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parse', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return input;
            }
            if (isArray(result)) {
                throw new ValidationError(result);
            }
            return result.value;
        };
        Object.defineProperty(this, 'parse', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    throw new ValidationError(result);
                }
                return result.value;
            });
        };
        Object.defineProperty(this, 'parseAsync', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseOrDefault', {
    get: function () {
        var _this = this;
        var cb = function (input, defaultValue, options) {
            var result = _this.apply(input, options || defaultParseOptions);
            if (result === null) {
                return input;
            }
            if (isArray(result)) {
                return defaultValue;
            }
            return result.value;
        };
        Object.defineProperty(this, 'parseOrDefault', { value: cb });
        return cb;
    },
});
Object.defineProperty(Shape.prototype, 'parseOrDefaultAsync', {
    get: function () {
        var _this = this;
        var cb = function (input, defaultValue, options) {
            return _this.applyAsync(input, options || defaultParseOptions).then(function (result) {
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    return defaultValue;
                }
                return result.value;
            });
        };
        Object.defineProperty(this, 'parseOrDefaultAsync', { value: cb });
        return cb;
    },
});
/**
 * The shape that applies a transformer callback to the base shape output.
 *
 * @template S The base shape.
 * @template O The transformed value.
 */
var TransformShape = /** @class */ (function (_super) {
    __extends(TransformShape, _super);
    /**
     * Creates the new {@linkcode TransformShape} instance.
     *
     * @param shape The base shape.
     * @param async If `true` then the transformed shape would await for the transformer to finish and use the resolved
     * value as an output. Otherwise, the value that is synchronously returned from the transformer is used as an output.
     * @param callback The callback that transforms the shape output value.
     * @template S The base shape.
     * @template O The transformed value.
     */
    function TransformShape(
    /**
     * The base shape which output value is transformed.
     */
    shape, async, 
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    callback) {
        var _this = _super.call(this, shape.inputTypes, shape.async || async) || this;
        _this.shape = shape;
        _this.callback = callback;
        return _this;
    }
    TransformShape.prototype.apply = function (input, options) {
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        var result = shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        try {
            output = callback(output, options);
        }
        catch (error) {
            return captureIssues(error);
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        if (isEqual(input, output)) {
            return null;
        }
        return ok(output);
    };
    TransformShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        return shape.applyAsync(input, options).then(function (result) {
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            return new Promise(function (resolve) { return resolve(callback(output, options)); }).then(function (output) {
                var issues;
                if (_applyChecks !== null) {
                    issues = _applyChecks(output, null, options);
                    if (issues !== null) {
                        return issues;
                    }
                }
                if (isEqual(input, output)) {
                    return null;
                }
                return ok(output);
            }, captureIssues);
        });
    };
    return TransformShape;
}(Shape));
export { TransformShape };
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template I The input shape.
 * @template O The output shape.
 */
var RedirectShape = /** @class */ (function (_super) {
    __extends(RedirectShape, _super);
    /**
     * Creates the new {@linkcode RedirectShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template I The input shape.
     * @template O The output shape.
     */
    function RedirectShape(
    /**
     * The input shape.
     */
    inputShape, 
    /**
     * The output shape.
     */
    outputShape) {
        var _this = _super.call(this, inputShape.inputTypes, inputShape.async || outputShape.async) || this;
        _this.inputShape = inputShape;
        _this.outputShape = outputShape;
        return _this;
    }
    RedirectShape.prototype.apply = function (input, options) {
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        var result = inputShape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        var outputResult = outputShape.apply(output, options);
        if (outputResult !== null) {
            if (isArray(outputResult)) {
                return outputResult;
            }
            result = outputResult;
            output = outputResult.value;
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    RedirectShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var result = null;
        var output = input;
        return inputShape
            .applyAsync(input, options)
            .then(function (inputResult) {
            if (inputResult !== null) {
                if (isArray(inputResult)) {
                    return inputResult;
                }
                result = inputResult;
                output = inputResult.value;
            }
            return outputShape.applyAsync(output, options);
        })
            .then(function (outputResult) {
            var issues;
            if (outputResult !== null) {
                if (isArray(outputResult)) {
                    return outputResult;
                }
                result = outputResult;
                output = outputResult.value;
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return RedirectShape;
}(Shape));
export { RedirectShape };
/**
 * The shape that replaces an input value with another value.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template A The replaced value.
 * @template B The replacement value.
 */
var ReplaceShape = /** @class */ (function (_super) {
    __extends(ReplaceShape, _super);
    /**
     * Creates the new {@linkcode ReplaceShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param searchedValue The input that should be replaced.
     * @param value The replacement value. If `undefined` then the replaced value is returned as is without any parsing.
     * @template S The shape that parses the input without the replaced value.
     * @template A The searched value.
     * @template B The replacement value.
     */
    function ReplaceShape(
    /**
     * The shape that parses the input without the replaced value.
     */
    shape, 
    /**
     * The replaced value.
     */
    searchedValue, 
    /**
     * The replacement value.
     */
    value) {
        var _this = _super.call(this, shape.inputTypes.concat(Shape.typeof(searchedValue)), shape.async) || this;
        _this.shape = shape;
        _this.searchedValue = searchedValue;
        _this.value = value;
        _this._replacedResult = value === undefined || isEqual(value, searchedValue) ? null : ok(value);
        return _this;
    }
    ReplaceShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var issues;
        var output = input;
        var result = isEqual(input, this.searchedValue) ? this._replacedResult : this.shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    ReplaceShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _applyChecks = this._applyChecks;
        var issues;
        if (isEqual(input, this.searchedValue)) {
            if (_applyChecks !== null) {
                return new Promise(function (resolve) {
                    issues = _applyChecks(_this.value, null, options);
                    resolve(issues !== null ? issues : _this._replacedResult);
                });
            }
            return (this._replacedResultPromise || (this._replacedResultPromise = Promise.resolve(this._replacedResult)));
        }
        return this.shape.applyAsync(input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return ReplaceShape;
}(Shape));
export { ReplaceShape };
/**
 * The shape that excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
var ExcludeShape = /** @class */ (function (_super) {
    __extends(ExcludeShape, _super);
    /**
     * Creates the new {@linkcode ExcludeShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param excludedValue The excluded value.
     * @param options The constraint options or an issue message.
     * @template S The shape that parses the input without the replaced value.
     * @template T The excluded value.
     */
    function ExcludeShape(
    /**
     * The base shape.
     */
    shape, 
    /**
     * The excluded value.
     */
    excludedValue, options) {
        var _this = _super.call(this, shape.inputTypes, shape.async) || this;
        _this.shape = shape;
        _this.excludedValue = excludedValue;
        _this._issueFactory = createIssueFactory(CODE_EXCLUSION, MESSAGE_EXCLUSION, options, excludedValue);
        return _this;
    }
    ExcludeShape.prototype.apply = function (input, options) {
        var _a = this, excludedValue = _a.excludedValue, _issueFactory = _a._issueFactory, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        if (isEqual(input, excludedValue)) {
            return [_issueFactory(input, options)];
        }
        var result = this.shape.apply(input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
            if (isEqual(output, excludedValue)) {
                return [_issueFactory(input, options)];
            }
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    ExcludeShape.prototype.applyAsync = function (input, options) {
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, excludedValue = _a.excludedValue, _issueFactory = _a._issueFactory, _applyChecks = _a._applyChecks;
        if (isEqual(input, excludedValue)) {
            return Promise.resolve([_issueFactory(input, options)]);
        }
        return this.shape.applyAsync(input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
                if (isEqual(output, excludedValue)) {
                    return [_issueFactory(input, options)];
                }
            }
            if (_applyChecks !== null) {
                issues = _applyChecks(output, null, options);
                if (issues !== null) {
                    return issues;
                }
            }
            return result;
        });
    };
    return ExcludeShape;
}(Shape));
export { ExcludeShape };
