import { __extends } from "tslib";
import { Shape } from './Shape';
import { appendCheck, createIssueFactory, stringTypes } from '../utils';
import { CODE_STRING_MAX, CODE_STRING_MIN, CODE_STRING_REGEX, CODE_TYPE, MESSAGE_STRING_MAX, MESSAGE_STRING_MIN, MESSAGE_STRING_REGEX, MESSAGE_STRING_TYPE, TYPE_STRING, } from '../constants';
/**
 * The shape that constrains the input as a string.
 */
var StringShape = /** @class */ (function (_super) {
    __extends(StringShape, _super);
    /**
     * Creates a new {@linkcode StringShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function StringShape(options) {
        var _this = _super.call(this, stringTypes) || this;
        _this._issueFactory = createIssueFactory(CODE_TYPE, MESSAGE_STRING_TYPE, options, TYPE_STRING);
        return _this;
    }
    /**
     * The shortcut to apply both {@linkcode min} and {@linkcode max} constraints.
     *
     * @param length The exact length a string must have.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the string length to be greater than or equal to the length.
     *
     * @param length The minimum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MIN, MESSAGE_STRING_MIN, options, length);
        return appendCheck(this, CODE_STRING_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string length to be less than or equal to the length.
     *
     * @param length The maximum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MAX, MESSAGE_STRING_MAX, options, length);
        return appendCheck(this, CODE_STRING_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string to match a regexp.
     *
     * @param re The regular expression that the sting must conform.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.regex = function (re, options) {
        var issueFactory = createIssueFactory(CODE_STRING_REGEX, MESSAGE_STRING_REGEX, options, re);
        return appendCheck(this, CODE_STRING_REGEX, options, re, function (input, options) {
            re.lastIndex = 0;
            if (!re.test(input)) {
                return issueFactory(input, options);
            }
        });
    };
    StringShape.prototype.apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (typeof input !== 'string') {
            return [this._issueFactory(input, options)];
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return StringShape;
}(Shape));
export { StringShape };
