import { __extends } from "tslib";
import { Shape } from './Shape';
import { concatIssues, createIssueFactory, getInputTypes, isArray, isAsyncShapes, unique } from '../utils';
import { CODE_UNION, MESSAGE_UNION } from '../constants';
/**
 * The shape that requires an input to conform at least one of the united shapes.
 *
 * @template U The list of united shapes.
 */
var UnionShape = /** @class */ (function (_super) {
    __extends(UnionShape, _super);
    /**
     * Creates a new {@linkcode UnionShape} instance.
     *
     * @param shapes The list of united shapes.
     * @param options The union constraint options or an issue message.
     * @template U The list of united shapes.
     */
    function UnionShape(
    /**
     * The list of united shapes.
     */
    shapes, options) {
        var _this = _super.call(this, getInputTypes(shapes), isAsyncShapes(shapes)) || this;
        _this.shapes = shapes;
        var _a = createUnionBuckets(shapes), buckets = _a.buckets, anyBucket = _a.anyBucket;
        _this._buckets = buckets;
        _this._anyBucket = anyBucket;
        _this._issueFactory = createIssueFactory(CODE_UNION, MESSAGE_UNION, options);
        return _this;
    }
    UnionShape.prototype.at = function (key) {
        var valueShapes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var valueShape = shape.at(key);
            if (valueShape !== null) {
                valueShapes.push(valueShape);
            }
        }
        if (valueShapes.length === 0) {
            return null;
        }
        if (valueShapes.length === 1) {
            return valueShapes[0];
        }
        return new UnionShape(valueShapes);
    };
    UnionShape.prototype.apply = function (input, options) {
        var _a = this, _buckets = _a._buckets, _anyBucket = _a._anyBucket, _applyChecks = _a._applyChecks;
        var bucket = _buckets !== null ? _buckets[Shape.typeof(input)] || _anyBucket : _anyBucket;
        var issues = null;
        var result = null;
        var output = input;
        var bucketLength = 0;
        var index = 0;
        if (bucket !== null) {
            for (bucketLength = bucket.length; index < bucketLength; ++index) {
                result = bucket[index].apply(input, options);
                if (result === null) {
                    break;
                }
                if (isArray(result)) {
                    issues = concatIssues(issues, result);
                    continue;
                }
                output = result.value;
                break;
            }
        }
        if (index === bucketLength) {
            return [this._issueFactory(input, options, issues)];
        }
        if (_applyChecks !== null) {
            issues = _applyChecks(output, null, options);
            if (issues !== null) {
                return issues;
            }
        }
        return result;
    };
    UnionShape.prototype.applyAsync = function (input, options) {
        var _this = this;
        if (!this.async) {
            return _super.prototype.applyAsync.call(this, input, options);
        }
        var _a = this, _buckets = _a._buckets, _anyBucket = _a._anyBucket, _applyChecks = _a._applyChecks;
        var bucket = _buckets !== null ? _buckets[Shape.typeof(input)] || _anyBucket : _anyBucket;
        if (bucket === null) {
            return Promise.resolve([this._issueFactory(input, options, [])]);
        }
        var bucketLength = bucket.length;
        var issues = null;
        var index = 0;
        var nextShape = function () {
            return bucket[index].applyAsync(input, options).then(function (result) {
                ++index;
                var output = input;
                if (result !== null) {
                    if (isArray(result)) {
                        issues = concatIssues(issues, result);
                        if (index === bucketLength) {
                            return [_this._issueFactory(input, options, issues)];
                        }
                        return nextShape();
                    }
                    output = result.value;
                }
                if (_applyChecks !== null) {
                    issues = _applyChecks(output, null, options);
                    if (issues !== null) {
                        return issues;
                    }
                }
                return result;
            });
        };
        return nextShape();
    };
    return UnionShape;
}(Shape));
export { UnionShape };
/**
 * Creates a mapping from the value type to an array of shapes that are applicable.
 *
 * @param shapes The list of united shapes.
 */
export function createUnionBuckets(shapes) {
    var buckets = {};
    var anyBucket = null;
    var bucketTypes = [];
    for (var _i = 0, _a = unwrapUnionShapes(shapes); _i < _a.length; _i++) {
        var shape = _a[_i];
        // Collect shapes that can parse any input
        if (shape.inputTypes.includes('any')) {
            anyBucket || (anyBucket = []);
            if (!anyBucket.includes(shape)) {
                anyBucket.push(shape);
            }
            continue;
        }
        // Populate buckets that require specific input types
        for (var _b = 0, _c = shape.inputTypes; _b < _c.length; _b++) {
            var type = _c[_b];
            var bucket = buckets[type];
            if (!bucket) {
                bucketTypes.push(type);
                buckets[type] = [shape];
                continue;
            }
            if (!bucket.includes(shape)) {
                bucket.push(shape);
            }
        }
    }
    if (bucketTypes.length === 0) {
        buckets = null;
    }
    else if (anyBucket !== null && buckets !== null) {
        for (var _d = 0, bucketTypes_1 = bucketTypes; _d < bucketTypes_1.length; _d++) {
            var type = bucketTypes_1[_d];
            buckets[type] = unique(buckets[type].concat(anyBucket));
        }
    }
    return { buckets: buckets, anyBucket: anyBucket };
}
/**
 * Unwraps nested union shapes that don't have any checks.
 */
function unwrapUnionShapes(opaqueShapes) {
    var shapes = [];
    for (var _i = 0, opaqueShapes_1 = opaqueShapes; _i < opaqueShapes_1.length; _i++) {
        var shape = opaqueShapes_1[_i];
        if (shape instanceof UnionShape && shape.checks.length === 0) {
            shapes.push.apply(shapes, unwrapUnionShapes(shape.shapes));
        }
        else {
            shapes.push(shape);
        }
    }
    return shapes;
}
